/* $Id: capimain.c,v 1.1.2.6 2001/02/16 08:40:36 armin Exp $
 *
 * ISDN interface module for Dialogic active cards DIVA.
 * CAPI Interface
 *
 * Copyright 2000-2007 by Armin Schindler (mac@melware.de)
 * Copyright 2000-2007 Cytronics & Melware (info@melware.de)
 * Copyright 2007 Dialogic
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define __KERNEL_SYSCALLS__
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/sched.h>

#include <linux/isdn.h>
#include <linux/isdnif.h>


#include "platform.h"
#include "di_defs.h"
#include "capi20.h"
#include "divacapi.h"
#include "divasync.h"
#include "cp_vers.h"
#include "drv_man.h"
#include "cfg_types.h"
#include "cfg_notify.h"
#include "dlist.h"

#include <linux/capi.h>
#include <linux/kernelcapi.h>
#include <../drivers/isdn/avmb1/capiutil.h>
#include <../drivers/isdn/avmb1/capilli.h>

EXPORT_NO_SYMBOLS;

#define DBG_MINIMUM  (DL_LOG + DL_FTL + DL_ERR)
#define DBG_DEFAULT  (DBG_MINIMUM + DL_XLOG + DL_REG)

static char *main_revision = "$Revision: 1.1.2.6 $";
static char *DRIVERNAME = "Dialogic DIVA - CAPI Interface driver (http://www.melware.net)";
static char *DRIVERLNAME = "divacapi";
int errno = 0;
#define DRRELMAJOR  3
#define DRRELMINOR  0
#define DRRELEXTRA  "3.0.10-107.884-1"
static char DRIVERRELEASE[16];
static int diva_capi_uses_madapter;
static int diva_handle_send_message_in_interrupt_context;

#define M_COMPANY "Dialogic"

typedef struct _diva_card {
	int Id;
	struct _diva_card *next;
	struct capi_ctr *capi_ctrl;
	DIVA_CAPI_ADAPTER *adapter;
	DESCRIPTOR d;
	char name[32];
	int PRI;
} diva_card;

typedef struct _diva_card_event_q {
	struct _diva_card_event_q* next;
	struct _diva_card*         card;
	struct sk_buff_head        inqueue;
} diva_card_event_q_t;

typedef struct _diva_os_thread_dpc {
  struct semaphore divas_sem;
  struct semaphore divas_end;
  int pid;
  atomic_t thread_running;
  diva_os_soft_isr_t* psoft_isr;
} diva_os_thread_dpc_t;

typedef struct _diva_mem_release {
  diva_entity_link_t link;
  unsigned long t;
  void* DataNCCI;
  void* DataFlags;
  void* ReceiveBuffer;
  void* xbuffer_internal;
  void* xbuffer_table;
  int   num_data;
  void** xbuffer_ptr;
} diva_mem_release_t;
static diva_entity_queue_t mem_release_q;

static struct {
	diva_os_soft_isr_t queue_soft_isr;
	spinlock_t event_q_lock;
} diva_capi_event_queue;

extern void AutomaticLaw(DIVA_CAPI_ADAPTER *);
extern void callback (ENTITY *);
extern word api_remove_start(void);
extern word CapiRelease(word);
extern word CapiRegister(word);
extern word api_put (APPL *, CAPI_MSG *);

byte UnMapController(byte);

static void diva_reset_ctr(struct capi_ctr *);
static void diva_remove_ctr(struct capi_ctr *);
static void diva_register_appl(struct capi_ctr *, __u16 , capi_register_params *);
static void diva_release_appl(struct capi_ctr *, __u16);
static void diva_send_message(struct capi_ctr *, struct sk_buff *);
static char *diva_procinfo(struct capi_ctr *);
static int diva_ctl_read_proc(char *, char **, off_t ,int , int *, struct capi_ctr *);
static int diva_driver_read_proc(char *, char **, off_t ,int , int *, struct capi_driver *);
static int diva_capi_read_cfg_value (IDI_SYNC_REQ* sync_req,
                                     int instance,
                                     const char* name,
                                     diva_cfg_lib_value_type_t value_type,
                                     void* data,
                                     dword data_length);
static void diva_process_mem_release_q(int force);
static void diva_init_event_queue(void);
static void diva_shutdown_event_queue(void);

void no_printf (unsigned char *, ...);

DIVA_DI_PRINTF dprintf = no_printf;

DIVA_CAPI_ADAPTER *adapter = (DIVA_CAPI_ADAPTER *) NULL;
APPL *application = (APPL *) NULL;
byte max_appl = MAX_APPL;
byte ControllerMap[MAX_ADAPTER + 1];
CAPI_MSG *mapped_msg = (CAPI_MSG *) NULL;
static byte max_adapter = 0;
struct resource_index_entry *c_ind_resource_index = (struct resource_index_entry *) NULL;

static void clean_adapter(int);
static void DIRequest (ENTITY* e);

MODULE_DESCRIPTION(             "CAPI driver for Dialogic DIVA cards");
MODULE_AUTHOR(                  "Cytronics & Melware, Dialogic");
MODULE_SUPPORTED_DEVICE(        "CAPI and DIVA card drivers");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


extern void DIVA_DIDD_Read(DESCRIPTOR *, int);

static dword notify_handle;
static DESCRIPTOR DAdapter;
static DESCRIPTOR MAdapter;

static diva_card*    cards;
static diva_card_event_q_t* events;

static struct capi_driver_interface *di;

static struct capi_driver divas_driver = {
	reset_ctr:diva_reset_ctr,
	remove_ctr:diva_remove_ctr,
	register_appl:diva_register_appl,
	release_appl:diva_release_appl,
	send_message:diva_send_message,
	procinfo:diva_procinfo,
	ctr_read_proc:diva_ctl_read_proc,
	driver_read_proc:diva_driver_read_proc,
	load_firmware:0,
	add_card:0
};

/************** not needed ***************/

static void diva_reset_ctr(struct capi_ctr *ctrl)
{
       /* not needed */
}

static void diva_remove_ctr(struct capi_ctr *ctrl)
{
       /* not needed */
}

static spinlock_t api_lock;
static spinlock_t ll_lock;
static spinlock_t mem_q_lock;

#include "debuglib.c"

void xlog (char * x, ...)
{
  va_list ap;

  if ( myDriverDebugHandle.dbgMask & DL_XLOG )
  {
    va_start (ap, x) ;

    if ( myDriverDebugHandle.dbg_irq )
    {
      myDriverDebugHandle.dbg_irq (myDriverDebugHandle.id, DLI_XLOG, x, ap);
    }
    else if ( myDriverDebugHandle.dbg_old )
    {
      myDriverDebugHandle.dbg_old (myDriverDebugHandle.id, x, ap) ;
    }

    va_end (ap) ;
  }

}

static char *
getrev(const char *revision)
{
	char *rev;
	char *p;
	if ((p = strchr(revision, ':'))) {
		rev = p + 2;
		p = strchr(rev, '$');
		*--p = 0;
	} else rev = "1.0";
	return rev;

}

static int
find_free_id(void)
{
  int num = 0;
  diva_card *p;

  spin_lock(&ll_lock);
  while(num < 100) {
    num++;
    p = cards;
    while (p) {
      if (p->Id == num)
        break;
      p = p->next;
    }
    if ((p) && (p->Id == num))
	    continue;
    spin_unlock(&ll_lock);
    return(num);
  }
  spin_unlock(&ll_lock);
  return(999);
}

static inline diva_card *
find_card_by_ctrl(word controller)
{
  diva_card *p;

  spin_lock(&ll_lock);
  p = cards;

  while (p) {
    if (ControllerMap[p->Id] == controller) {
      spin_unlock(&ll_lock);
      return p;
    }
    p = p->next;
  }
  spin_unlock(&ll_lock);
  return (diva_card *) 0;
}



/************* interface functions *************/

void no_printf (unsigned char * x ,...)
{
  /* dummy debug function */
}

/*
** main function called by message.c
*/
void sendf(APPL *appl, word command, dword Id, word Number, byte *format, ...)
{
  word i, j;
  word length = 12, dlength = 0;
  byte *write;
  CAPI_MSG msg;
  byte *string = 0;
  va_list ap;
  struct sk_buff *skb;
  diva_card *card = NULL;
  dword tmp;

  if(!appl) return;

  DBG_PRV1(("sendf(a=%d,cmd=%x,format=%s)",
             appl->Id,command,(byte *)format))

  msg.header.appl_id = appl->Id;
  msg.header.command = command;
  if((byte)(command>>8) == 0x82)
    Number = appl->Number++;
  msg.header.number = Number;

  WRITE_DWORD(((byte *)&msg.header.controller), Id);
  write = (byte *) &msg;
  write += 12;

  va_start(ap,format);
  for(i=0; format[i]; i++) {
    switch(format[i]) {
      case 'b':
        tmp = va_arg(ap,dword);
        *(byte *)write = (byte)(tmp & 0xff);
        write += 1;
        length += 1;
        break;
      case 'w':
        tmp = va_arg(ap,dword);
        WRITE_WORD (write, (tmp & 0xffff));
        write += 2;
        length += 2;
        break;
      case 'd':
        tmp = va_arg(ap,dword);
        WRITE_DWORD(write, tmp);
        write += 4;
        length += 4;
        break;
      case 's':
      case 'S':
        string = va_arg(ap,byte *);
        length += string[0]+1;
        for(j = 0; j <= string[0]; j++)
          *write++ = string[j];
        break;
    }
  }
  va_end(ap);

  msg.header.length = length;
  msg.header.controller = UnMapController(msg.header.controller);

  if (command == _DATA_B3_I)
    dlength = READ_WORD(((byte*)&msg.info.data_b3_ind.Data_Length));

  if (!(skb = alloc_skb(length + dlength, GFP_ATOMIC))) {
    printk(KERN_ERR "%s: alloc_skb failed, incoming msg dropped.\n", DRIVERLNAME);
    return;
  } else {
    write = (byte *) skb_put(skb, length + dlength);

    /* copy msg header to sk_buff */
    memcpy(write, (byte *)&msg, length);

    /* if DATA_B3_IND, copy data too */
    if (command == _DATA_B3_I) {
      dword data = READ_DWORD(((byte*)&msg.info.data_b3_ind.Data));
      memcpy (write + length, (void *)data, dlength);
    }

    if ( myDriverDebugHandle.dbgMask & DL_XLOG )
    {
      switch ( command )
      {
        default:
          xlog ("\x00\x02", &msg, 0x81, length) ;
          break ;

        case _DATA_B3_R|CONFIRM:
          if ( myDriverDebugHandle.dbgMask & DL_BLK)
            xlog ("\x00\x02", &msg, 0x81, length) ;
          break ;

        case _DATA_B3_I:
          if ( myDriverDebugHandle.dbgMask & DL_BLK)
          {
            xlog ("\x00\x02", &msg, 0x81, length) ;
            for (i = 0; i < dlength; i += 256)
            {
              DBG_BLK((((char *)msg.info.data_b3_ind.Data) + i,
                ((dlength - i) < 256) ?
                 (dlength - i) : 256  ))
              if ( !(myDriverDebugHandle.dbgMask & DL_PRV0) )
                break ; // not more if not explicitely requested
            }
          }
          break ;
      }
    }

    /* find the card structure for this controller */
    if(!(card = find_card_by_ctrl(skb->data[8] & 0x7f))) {
      printk(KERN_ERR "%s: controller %d not found, incoming msg dropped.\n",
              DRIVERLNAME, skb->data[8] & 0x7f);
      DBG_ERR(("sendf - controller not found, incoming msg dropped"))
      kfree_skb(skb);
      return;
    }

    /* tell capi that we have a new ncci assigned */
    if ((command == (_CONNECT_B3_R | CONFIRM)) ||
        (command == _CONNECT_B3_I)) {
      card->capi_ctrl->new_ncci(card->capi_ctrl, appl->Id,
				CAPIMSG_NCCI(skb->data), MAX_DATA_B3);
    }
    /* tell capi that we have released the ncci */
    if (command == _DISCONNECT_B3_I) {
      card->capi_ctrl->free_ncci(card->capi_ctrl, appl->Id, CAPIMSG_NCCI(skb->data));
    }

    /* send capi msg to capi */
    card->capi_ctrl->handle_capimsg(card->capi_ctrl, appl->Id, skb);
  }
}

void sync_callback(ENTITY *e)
{
  DBG_TRC(("cb:Id=%x,Rc=%x,Ind=%x", e->Id, e->Rc, e->Ind))

  spin_lock(&api_lock);

  callback(e);

  spin_unlock(&api_lock);
}


/*
** TX Buffer allocation
*/

#define XBUFFER_USED_ANCHOR   0
#define XBUFFER_FREE_ANCHOR   1
#define XBUFFER_ANCHOR_COUNT  2

#define XBufferNext(t) ((t)[XBUFFER_FREE_ANCHOR].next_i - XBUFFER_ANCHOR_COUNT)

void XBufferInit (struct xbuffer_element *t, word n)
{
  word i;

  t[XBUFFER_USED_ANCHOR].prev_i = XBUFFER_USED_ANCHOR;
  t[XBUFFER_USED_ANCHOR].next_i = XBUFFER_USED_ANCHOR;
  t[XBUFFER_FREE_ANCHOR].prev_i = XBUFFER_FREE_ANCHOR;
  t[XBUFFER_FREE_ANCHOR].next_i = XBUFFER_FREE_ANCHOR;
  for (i = 0; i < n; i++)
  {
    t[XBUFFER_ANCHOR_COUNT + i].prev_i = t[XBUFFER_FREE_ANCHOR].prev_i;
    t[XBUFFER_ANCHOR_COUNT + i].next_i = XBUFFER_FREE_ANCHOR;
    t[t[XBUFFER_FREE_ANCHOR].prev_i].next_i = XBUFFER_ANCHOR_COUNT + i;
    t[XBUFFER_FREE_ANCHOR].prev_i = XBUFFER_ANCHOR_COUNT + i;
  }
}

void XBufferAlloc (struct xbuffer_element *t, word e)
{
  word i;

  i = e + XBUFFER_ANCHOR_COUNT;
  t[t[i].prev_i].next_i = t[i].next_i;
  t[t[i].next_i].prev_i = t[i].prev_i;
  t[i].next_i = t[XBUFFER_USED_ANCHOR].next_i;
  t[i].prev_i = XBUFFER_USED_ANCHOR;
  t[t[XBUFFER_USED_ANCHOR].next_i].prev_i = i;
  t[XBUFFER_USED_ANCHOR].next_i = i;
}

void XBufferFree (struct xbuffer_element *t, word e)
{
  word i;

  i = e + XBUFFER_ANCHOR_COUNT;
  t[t[i].prev_i].next_i = t[i].next_i;
  t[t[i].next_i].prev_i = t[i].prev_i;
  t[i].next_i = t[XBUFFER_FREE_ANCHOR].next_i;
  t[i].prev_i = XBUFFER_FREE_ANCHOR;
  t[t[XBUFFER_FREE_ANCHOR].next_i].prev_i = i;
  t[XBUFFER_FREE_ANCHOR].next_i = i;
}


/*
** Buffer RX/TX
*/
void *TransmitBufferSet(APPL *appl, dword ref)
{
  XBufferAlloc (appl->xbuffer_table, (word) ref);
  DBG_PRV1(("%d:xbuf_used(%d)", appl->Id, ref + 1))
  return (void *) ref;
}

void *TransmitBufferGet(APPL *appl, void *p)
{
  if(appl->xbuffer_internal[(dword)p])
    return appl->xbuffer_internal[(dword)p];

  return appl->xbuffer_ptr[(dword)p];
}

void TransmitBufferFree(APPL *appl, void *p)
{
  XBufferFree (appl->xbuffer_table, (word)((dword) p));
  DBG_PRV1(("%d:xbuf_free(%d)", appl->Id, ((dword) p) + 1))
}

void *ReceiveBufferGet(APPL *appl, int Num)
{
  return &appl->ReceiveBuffer[Num * appl->MaxDataLength];
}


/*
** api_remove_start/complete for cleanup
*/

void api_remove_complete(void)
{
    DBG_PRV1(("api_remove_complete"))
}


/*
** Controller mapping
*/

byte MapController(byte Controller)
{
  byte i;
  byte MappedController = 0;
  byte ctrl = Controller & 0x7f;  // mask external controller bit off

  for (i = 1; i < max_adapter + 1; i++)
  {
    if (ctrl == ControllerMap[i])
    {
      MappedController = (byte) i;
      break;
    }
  }
  if (i > max_adapter)
  {
    ControllerMap[0] = ctrl;
    MappedController = 0;
  }
  return (MappedController | (Controller & 0x80)); // put back external controller bit
}

byte UnMapController(byte MappedController)
{
  byte Controller;
  byte ctrl = MappedController & 0x7f;  // mask external controller bit off

  if (ctrl <= max_adapter)
  {
    Controller = ControllerMap[ctrl];
  }
  else
  {
    Controller = 0;
  }

  return (Controller | (MappedController & 0x80)); // put back external controller bit
}


/*
** we do not provide date/time here,
** the application should do this.
*/
int fax_head_line_time (char *buffer)
{
  return (0);
}


/************* proc functions *************/

static int diva_driver_read_proc(char *page, char **start, off_t off,int count, int *eof, struct capi_driver *driver)
{
  int len = 0;
  char tmprev[32];

  strcpy(tmprev, main_revision);
  len += sprintf(page+len, "%-16s divas\n", "name");
  len += sprintf(page+len, "%-16s %s/%s %s(%s)\n", "release",
                 DRIVERRELEASE, getrev(tmprev), diva_capi_common_code_build, DIVA_BUILD);
  len += sprintf(page+len, "%-16s %s\n", "author", "Cytronics & Melware / Dialogic");

  if (off + count >= len)
    *eof = 1;
  if (len < off)
    return 0;
  *start = page + off;
  return((count < len-off) ? count : len-off);
}

static int diva_ctl_read_proc(char *page, char **start, off_t off,int count, int *eof, struct capi_ctr *ctrl)
{
  diva_card *card = ctrl->driverdata;
  int len = 0;

  len += sprintf(page+len, "%s\n", ctrl->name);
  len += sprintf(page+len, "Serial No. : %s\n", ctrl->serial);
  len += sprintf(page+len, "Id         : %d\n", card->Id);
  len += sprintf(page+len, "Channels   : %d\n", card->d.channels);

  if (off + count >= len)
    *eof = 1;
  if (len < off)
    return 0;
  *start = page + off;
  return((count < len-off) ? count : len-off);
}

static char *diva_procinfo(struct capi_ctr *ctrl)
{
  return(ctrl->serial);
}


/************** capi functions ***********/

/*
**  register appl
*/

static void diva_register_appl(struct capi_ctr *ctrl, __u16 appl, capi_register_params *rp)
{
  APPL *this;
  word bnum, xnum;
  int i = 0, j = 0;
  void * DataNCCI, * DataFlags, * ReceiveBuffer;
  struct xbuffer_element * xbuffer_table;
  void ** xbuffer_ptr, ** xbuffer_internal;

  MOD_INC_USE_COUNT;

  if(in_interrupt()) {
    DBG_ERR(("CAPI_REGISTER - in interrupt context !"))
    printk(KERN_ERR "%s: diva_register_appl: in interrupt context.\n", DRIVERLNAME);
    return;
  }

  diva_process_mem_release_q(0);

  DBG_TRC(("application register"))

  if(appl > MAX_APPL) {
    DBG_ERR(("CAPI_REGISTER - appl.Id exceeds MAX_APPL"))
    printk(KERN_WARNING "%s: diva_register_appl: appl.Id exceeds MAX_APPL.\n", DRIVERLNAME);
    return;
  }

  if ((rp->level3cnt < 1) ||
      (rp->datablkcnt < 1) ||
      (rp->datablklen < 80) ||
      (rp->datablklen > 2150) ||
      (((dword)(((long) rp->level3cnt) * ((long) rp->datablkcnt))) >= 0xffffL) ||
      (((dword)(((long) rp->level3cnt) * ((long) MAX_DATA_B3))) >= 0xffffL))
  {
    DBG_ERR(("CAPI_REGISTER - invalid parameters"))
    printk(KERN_WARNING "%s: diva_register_appl: invalid parameters.\n", DRIVERLNAME);
    return;
  }

  if (application[appl - 1].Id == appl) {
    ctrl->appl_registered(ctrl, appl);
    return; /* appl already registered */
  }

  /* alloc memory */

  bnum = rp->level3cnt * rp->datablkcnt;
  xnum = rp->level3cnt * MAX_DATA_B3;

  //     this->queue = ExAllocatePool (NonPagedPool, this->queue_size);

  if(!(DataNCCI = kmalloc(bnum * sizeof(word), GFP_ATOMIC))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc DataNCCI.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    return;
  }
  memset(DataNCCI, 0, bnum * sizeof(word));

  if(!(DataFlags = kmalloc(bnum * sizeof(word), GFP_ATOMIC))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc DataFlags.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    kfree(DataNCCI);
    return;
  }
  memset(DataFlags, 0, bnum * sizeof(word));

  if(!(ReceiveBuffer = vmalloc(bnum * rp->datablklen))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc ReceiveBuffer.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    kfree(DataNCCI);
    kfree(DataFlags);
    return;
  }
  memset(ReceiveBuffer, 0, bnum * rp->datablklen);

  if(!(xbuffer_table = (struct xbuffer_element *) kmalloc((XBUFFER_ANCHOR_COUNT +
       xnum) * sizeof(struct xbuffer_element), GFP_ATOMIC))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc xbuffer_table.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    kfree(DataNCCI);
    kfree(DataFlags);
    vfree(ReceiveBuffer);
    return;
  }
  XBufferInit (xbuffer_table, xnum);

  if(!(xbuffer_ptr = (void **) kmalloc(xnum * sizeof(void *), GFP_ATOMIC))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc xbuffer_ptr.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    kfree(DataNCCI);
    kfree(DataFlags);
    vfree(ReceiveBuffer);
    kfree(xbuffer_table);
    return;
  }
  memset(xbuffer_ptr, 0, xnum * sizeof(void *));

  if(!(xbuffer_internal = (void **) kmalloc(xnum * sizeof(void *), GFP_ATOMIC))) {
    printk(KERN_WARNING "%s: diva_register_appl: failed alloc xbuffer_internal.\n", DRIVERLNAME);
    DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
    kfree(DataNCCI);
    kfree(DataFlags);
    vfree(ReceiveBuffer);
    kfree(xbuffer_table);
    kfree(xbuffer_ptr);
    return;
  }
  memset(xbuffer_internal, 0, xnum * sizeof(void *));

  for ( i = 0 ; i < xnum ; i++ )
  {
    xbuffer_ptr[i] = kmalloc(rp->datablklen, GFP_ATOMIC) ;
    if ( !xbuffer_ptr[i] )
    {
      printk(KERN_WARNING "%s: diva_register_appl: failed alloc xbuffer_ptr block.\n", DRIVERLNAME);
      DBG_ERR(("CAPI_REGISTER - memory allocation failed"))
      if (i) {
        for(j = 0; j < i; j++)
          if(xbuffer_ptr[j])
            kfree(xbuffer_ptr[j]);
      }
      kfree(DataNCCI);
      kfree(DataFlags);
      vfree(ReceiveBuffer);
      kfree(xbuffer_table);
      kfree(xbuffer_ptr);
      kfree(xbuffer_internal);
      return;
    }
  }

  DBG_LOG(("CAPI_REGISTER - Id = %d", appl))
  DBG_LOG(("  MaxLogicalConnections = %d", rp->level3cnt))
  DBG_LOG(("  MaxBDataBuffers       = %d", rp->datablkcnt))
  DBG_LOG(("  MaxBDataLength        = %d", rp->datablklen))

  /* initialize application data */

  spin_lock(&api_lock);

  this = &application[appl - 1] ;
  memset(this, 0, sizeof (APPL));

  this->Id = appl;

  /* We do not need a list */
  /* InitializeListHead(&this->s_function); */

  for (i = 0; i < max_adapter; i++)
  {
    adapter[i].CIP_Mask[appl - 1] = 0;
  }

  this->queue_size = 1000;

  this->MaxNCCI = rp->level3cnt;
  this->MaxNCCIData = rp->datablkcnt;
  this->MaxBuffer = bnum;
  this->MaxDataLength = rp->datablklen;

  this->DataNCCI         = DataNCCI;
  this->DataFlags        = DataFlags;
  this->ReceiveBuffer    = ReceiveBuffer;
  this->xbuffer_table    = xbuffer_table;
  this->xbuffer_ptr      = xbuffer_ptr;
  this->xbuffer_internal = xbuffer_internal;
  for ( i = 0 ; i < xnum ; i++ ) {
    this->xbuffer_ptr[i] = xbuffer_ptr[i];
  }

  CapiRegister(this->Id);
  spin_unlock(&api_lock);

  /* tell kcapi that we actually registerd the appl */
  ctrl->appl_registered(ctrl, appl);
}


/*
**  release appl
*/

static void diva_release_appl(struct capi_ctr *ctrl, __u16 appl)
{
  APPL *this = &application[appl - 1] ;
  int i = 0;
  void* DataNCCI = 0;
  void* DataFlags = 0;
  void* ReceiveBuffer = 0;
  void** xbuffer_ptr = 0;
  void* xbuffer_internal = 0;
  void* xbuffer_table = 0;
  diva_mem_release_t* mem_release = 0;
  word MaxNCCI = 0;

  if(in_interrupt()) {
    DBG_ERR(("CAPI_RELEASE - in interrupt context !"))
    printk(KERN_ERR "%s: diva_release_appl: in interrupt context.\n", DRIVERLNAME);
    return;
  }

  diva_process_mem_release_q(0);

  if (this->Id) {
    DBG_TRC(("application %d cleanup", this->Id))
  }

  spin_lock(&api_lock);
  if (this->Id)
  {
    CapiRelease(this->Id);

    DataNCCI = this->DataNCCI;
    this->DataNCCI = 0;
    DataFlags = this->DataFlags;
    this->DataFlags = 0;
    ReceiveBuffer = this->ReceiveBuffer; /* vfree */
    this->ReceiveBuffer = 0;
    MaxNCCI = this->MaxNCCI;
    xbuffer_ptr = this->xbuffer_ptr;
    this->xbuffer_ptr = 0;
    xbuffer_internal = this->xbuffer_internal;
    this->xbuffer_internal = 0;
    xbuffer_table = this->xbuffer_table;
    this->xbuffer_table = 0;
    this->Id = 0;
  }
  spin_unlock(&api_lock);

  if ((mem_release = kmalloc(sizeof(*mem_release), GFP_ATOMIC)) != 0)
  {
    memset(mem_release, 0x00, sizeof(*mem_release));
    mem_release->t         = jiffies;
    mem_release->DataNCCI = DataNCCI;
    mem_release->DataFlags = DataFlags;
    mem_release->ReceiveBuffer = ReceiveBuffer;
    mem_release->num_data = MAX_DATA_B3 * MaxNCCI;
    mem_release->xbuffer_ptr = xbuffer_ptr;
    mem_release->xbuffer_internal = xbuffer_internal;
    mem_release->xbuffer_table = xbuffer_table;

    spin_lock(&mem_q_lock);
    diva_q_add_tail (&mem_release_q, &mem_release->link);
    spin_unlock(&mem_q_lock);
  }
  else
  {
    if (DataNCCI)
      kfree(DataNCCI);
    if (DataFlags)
      kfree(DataFlags);
    if (ReceiveBuffer)
      vfree(ReceiveBuffer);
    if (xbuffer_ptr) {
      for (i = 0; i < (MAX_DATA_B3 * MaxNCCI); i++)
        if (xbuffer_ptr[i]) kfree(xbuffer_ptr[i]);
      kfree(xbuffer_ptr);
    }
    if (xbuffer_internal)
      kfree(xbuffer_internal);
    if (xbuffer_table)
      kfree(xbuffer_table);
  }


  /* tell kcapi that we actually released the appl */
  ctrl->appl_released(ctrl, appl);
  MOD_DEC_USE_COUNT;
}


/*
**      send message
*/

static void diva_internal_send_message(struct capi_ctr *ctrl, struct sk_buff *skb)
{
  int i = 0, j = 0;
  word ret = 0;
  CAPI_MSG *msg = (CAPI_MSG *)skb->data;
  APPL *this;
  diva_card *card = ctrl->driverdata;
  __u32 length = skb->len;

  if(in_interrupt()) {
    DBG_ERR(("CAPI_SEND_MSG - in interrupt context !"))
    printk(KERN_ERR "%s: diva_send_message: in interrupt context.\n", DRIVERLNAME);
		kfree_skb(skb);
    return;
  }

  spin_lock(&api_lock);

  if ((msg->header.appl_id - 1) >= MAX_APPL) {
		DBG_ERR(("Write - wrong appl Id = %d, cmd = 0x%x", msg->header.appl_id - 1, msg->header.command))
    goto write_end;
  }
  this = &application[msg->header.appl_id - 1];
  if (this->Id == 0) {
		DBG_ERR(("Write - wrong appl = %d, cmd = 0x%x", msg->header.appl_id - 1, msg->header.command))
    goto write_end;
  }

	DBG_PRV1(("Write - appl = %d, cmd = 0x%x", this->Id, msg->header.command))

	/* patch controller number */
	msg->header.controller = ControllerMap[card->Id]
                           | (msg->header.controller &0x80);  /* preserve external controller bit */

  switch ( msg->header.command )
  {
    default:
      xlog ("\x00\x02", msg, 0x80, msg->header.length) ;
      break;

    case _DATA_B3_I|RESPONSE:
      if ( myDriverDebugHandle.dbgMask & DL_BLK )
        xlog ("\x00\x02", msg, 0x80, msg->header.length) ;
      break;

    case _DATA_B3_R:
      if ( myDriverDebugHandle.dbgMask & DL_BLK )
        xlog ("\x00\x02", msg, 0x80, msg->header.length) ;

      if (msg->header.length == 24) msg->header.length = 22; /* workaround for PPcom bug */
                                                             /* header is always 22      */
      if (msg->info.data_b3_req.Data_Length > this->MaxDataLength) 
      {
        DBG_ERR(("Write - invalid message size %d, maximum: %d", msg->info.data_b3_req.Data_Length, this->MaxDataLength))
        goto write_end;
      }

      if (msg->info.data_b3_req.Data_Length > (length - msg->header.length)) 
      {
        DBG_ERR(("Write - invalid message size %d, expected: %d", length, msg->header.length + msg->info.data_b3_req.Data_Length))
        goto write_end;
      }

      if ((i = XBufferNext (this->xbuffer_table)) < 0)
      {
        DBG_ERR(("Write - too many data pending"))
        goto write_end;
      }
      msg->info.data_b3_req.Data = i;

      this->xbuffer_internal[i] = NULL;
      memcpy (this->xbuffer_ptr[i], &((__u8 *)msg)[msg->header.length], msg->info.data_b3_req.Data_Length);

      if ( (myDriverDebugHandle.dbgMask & DL_BLK)
        && (myDriverDebugHandle.dbgMask & DL_XLOG))
      {
        for ( j = 0 ; j < msg->info.data_b3_req.Data_Length ; j += 256 )
        {
          DBG_BLK((((char *)this->xbuffer_ptr[i]) + j,
            ((msg->info.data_b3_req.Data_Length - j) < 256)?
             (msg->info.data_b3_req.Data_Length - j) : 256  ))
          if ( !(myDriverDebugHandle.dbgMask & DL_PRV0) )
            break ; // not more if not explicitely requested
        }
      }
      break ;
  }

  memcpy (mapped_msg, msg, (__u32)msg->header.length);
  mapped_msg->header.controller = MapController (mapped_msg->header.controller);

  ret = api_put (this, mapped_msg);
  switch (ret)
  {
    case 0:
      break;
    case _BAD_MSG:
      DBG_ERR(("Write - bad message"))
      break;
    case _QUEUE_FULL:
      DBG_ERR(("Write - queue full"))
      break;
    default:
      DBG_ERR(("Write - api_put returned unknown error"))
      break;
  }

write_end:
  spin_unlock(&api_lock);
  kfree_skb(skb);
}

static void diva_send_message(struct capi_ctr *ctrl, struct sk_buff *skb) {
	if (diva_handle_send_message_in_interrupt_context == 0) {
		diva_internal_send_message(ctrl, skb);
	} else {
		diva_card_event_q_t* event;
		unsigned long flags;

		spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);
		event = events;
		while (event) {
			if (event->card->capi_ctrl == ctrl) {
				skb_queue_tail(&event->inqueue, skb);
				spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
				diva_os_schedule_soft_isr (&diva_capi_event_queue.queue_soft_isr);
				return;
			}
			event = event->next;
		}
		spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
		kfree_skb(skb);
	}
}

static void
DIRequest (ENTITY* e)
{
  DIVA_CAPI_ADAPTER *a = &(adapter[(byte)e->user[0]]);
  diva_card* os_card = (diva_card*)a->os_card;

  if (e->Req && (a->FlowControlIdTable[e->ReqCh] == e->Id)) {
    a->FlowControlSkipTable[e->ReqCh] = 1;
  }

  (*(os_card->d.request))(e);
}

/************ module functions ***********/

static void __exit
diva_remove_cards(void)
{
  diva_card *last;
  diva_card *card;
	diva_card_event_q_t* local_events = 0, *last_event;
	struct sk_buff *skb;

  spin_lock(&ll_lock);
  card = cards;

  while(card) {
    di->detach_ctr(card->capi_ctrl);
    clean_adapter(card->Id - 1);
    DBG_TRC(("adapter remove, max_adapter=%d", max_adapter));
    card = card->next;
  }

	if (diva_handle_send_message_in_interrupt_context != 0) {
		unsigned long flags;
		spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);
		local_events = events;
		events       = 0;
		spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
	}

  card = cards;
  while (card) {
    last = card;
    card = card->next;
		kfree(last);
  }
  spin_unlock(&ll_lock);

	while (local_events) {
		last_event = local_events;
		local_events = local_events->next;
		while ((skb = skb_dequeue(&last_event->inqueue))) {
			kfree_skb(skb);
		}
		kfree(last_event);
	}
}

static void
diva_remove_card(DESCRIPTOR *d)
{
  diva_card *last;
  diva_card *card;
	diva_card_event_q_t* event = 0, *last_event;

  spin_lock(&ll_lock);

	if (diva_handle_send_message_in_interrupt_context != 0) {
		unsigned long flags;
		spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);
		last_event = event = events;
		while (event) {
			if (event->card->d.request == d->request) {
				if (event == last_event)
					events = event->next;
				else
					last_event->next = event->next;
				break;
			}
			last_event = event;
			event = event->next;
		}
		spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
	}

  last = card = cards;
  while(card) {
    if (card->d.request == d->request)
    {
      di->detach_ctr(card->capi_ctrl);
      clean_adapter(card->Id - 1);
      DBG_TRC(("DelAdapterMap (%d) -> (%d)", ControllerMap[card->Id], card->Id))
      ControllerMap[card->Id] = 0;
      DBG_TRC(("adapter remove, max_adapter=%d", max_adapter));
      if (card == last)
        cards = card->next;
      else
        last->next = card->next;

      kfree(card);
      break;
    }
    last = card;
    card = card->next;
  }
  spin_unlock(&ll_lock);

	if (event != 0) {
		struct sk_buff *skb;

		while ((skb = skb_dequeue(&event->inqueue))) {
			kfree_skb(skb);
		}

		kfree(event);
	}
}

static int
diva_add_card(DESCRIPTOR *d)
{
  int k = 0, i = 0;
  diva_card *card = NULL;
  struct capi_ctr *ctrl = NULL;
  DIVA_CAPI_ADAPTER *a = NULL;
  IDI_SYNC_REQ sync_req;
  char serial[16];
#if IMPLEMENT_LINE_INTERCONNECT2
  LI_CONFIG *new_li_config_table;
  int j;
  unsigned long table_size;
  byte *cp;
#endif /* IMPLEMENT_LINE_INTERCONNECT2 */
  diva_card_event_q_t* event = 0;
  unsigned long flags;
  dword serial_nr = 0;
  dword cardtype = 0;
  byte xdi_logical_adapter_number = 0;
  byte vscapi_mode = 0;
  const struct _diva_mtpx_adapter_xdi_config_entry* mtpx_cfg = 0;
  dword mtpx_cfg_number_of_configuration_entries = 0;

  if (!(card = (diva_card *) kmalloc(sizeof(diva_card), GFP_KERNEL))) {
    printk(KERN_ERR "%s: failed to allocate card struct.\n", DRIVERLNAME);
    return(0);
  }
	if (diva_handle_send_message_in_interrupt_context) {
		if (!(event = (diva_card_event_q_t*)kmalloc(sizeof(*event), GFP_KERNEL))) {
			kfree(card);
			printk(KERN_ERR "%s: failed to allocate card event struct.\n", DRIVERLNAME);
			return(0);
		}

		event->card = card;
		skb_queue_head_init(&event->inqueue);
	}


  memset((char *)card, 0x00, sizeof(diva_card));
  memcpy(&card->d, d, sizeof(DESCRIPTOR));

  if (diva_capi_uses_madapter) {
    card->d.type = IDI_ADAPTER_MAESTRA;
  }

  /*
    Retrieve channel count from management interface of XDI adapter
    */
  {
    diva_mgnt_adapter_info_t diva_adapter_info;
    int channels = diva_capi_read_channel_count (d, &diva_adapter_info);
    if (channels > 0) {
      card->d.channels = (byte)channels;
      cardtype = diva_adapter_info.cardtype;
      serial_nr = diva_adapter_info.serial_number;
      xdi_logical_adapter_number = (byte)diva_adapter_info.logical_adapter_number;
      vscapi_mode = (diva_adapter_info.vscapi_mode != 0);
      mtpx_cfg_number_of_configuration_entries = diva_adapter_info.mtpx_cfg_number_of_configuration_entries;
      mtpx_cfg = diva_adapter_info.mtpx_cfg;
    }
  }

  sync_req.GetName.Req = 0;
  sync_req.GetName.Rc = IDI_SYNC_REQ_GET_NAME;
  card->d.request((ENTITY *)&sync_req);
  strncpy(card->name, sync_req.GetName.name, sizeof(card->name));

  if (!(card->capi_ctrl = di->attach_ctr(&divas_driver, card->name, card))) {
    printk(KERN_ERR "%s: failed to attach controller.\n", DRIVERLNAME);
    kfree(card);
    if (event) kfree(event);
		if (mtpx_cfg != 0) diva_os_free (0, mtpx_cfg);
    return(0);
  }

  card->PRI = (card->d.channels > 2);
  if (cardtype == 0)
  {
    sync_req.GetCardType.Req = 0;
    sync_req.GetCardType.Rc  = IDI_SYNC_REQ_GET_CARDTYPE;
    sync_req.GetCardType.cardtype = 0;
    card->d.request((ENTITY *)&sync_req);
    cardtype = sync_req.GetCardType.cardtype;
  }
  switch (cardtype)
  {
  /* from channel count incorrectly classified PRI type adapters */
  case 92 /* CARDTYPE_DIVASRV_ANALOG_2PORT */:
  case 93 /* CARDTYPE_DIVASRV_V_ANALOG_2PORT */:
    card->PRI = TRUE;
  }
  card->Id = find_free_id();
  ctrl = card->capi_ctrl;
  strncpy(ctrl->manu, M_COMPANY, CAPI_MANUFACTURER_LEN);
  ctrl->version.majorversion = 2;
  ctrl->version.minorversion = 0;
  ctrl->version.majormanuversion = DRRELMAJOR;
  ctrl->version.minormanuversion = DRRELMINOR;
  if (serial_nr != 0) {
    sync_req.GetSerial.serial = serial_nr;
  } else {
    sync_req.GetSerial.Req = 0;
    sync_req.GetSerial.Rc = IDI_SYNC_REQ_GET_SERIAL;
    sync_req.GetSerial.serial = 0;
    card->d.request((ENTITY *)&sync_req);
  }
  if ((i = ((sync_req.GetSerial.serial & 0xff000000) >> 24))) {
    sprintf(serial, "%ld-%d",
            sync_req.GetSerial.serial & 0x00ffffff, i + 1);
  } else {
    sprintf(serial, "%ld", sync_req.GetSerial.serial);
  }
  serial[CAPI_SERIAL_LEN-1] = 0;
  strncpy(ctrl->serial, serial, CAPI_SERIAL_LEN);

  DBG_LOG(("Add IDI adapter Ch:%d, SN:'%s', Type:%ld",
           card->d.channels, serial, (long) cardtype))

  a = &adapter[card->Id - 1];
  a->mtpx_cfg.cfg = mtpx_cfg;
  a->mtpx_cfg.number_of_configuration_entries = mtpx_cfg_number_of_configuration_entries;
  card->adapter = a;
  a->os_card = card;
  ControllerMap[card->Id] = (byte)(ctrl->cnr);

  DBG_LOG(("AddAdapterMap (%d) -> (%d)", ctrl->cnr, card->Id))

  sync_req.xdi_capi_prms.Req = 0;
  sync_req.xdi_capi_prms.Rc = IDI_SYNC_REQ_XDI_GET_CAPI_PARAMS;
  sync_req.xdi_capi_prms.info.structure_length = sizeof(diva_xdi_get_capi_parameters_t);
  card->d.request((ENTITY *)&sync_req);
  a->flag_dynamic_l1_down = sync_req.xdi_capi_prms.info.flag_dynamic_l1_down;
  a->group_optimization_enabled = sync_req.xdi_capi_prms.info.group_optimization_enabled;

  a->xdi_logical_adapter_number = xdi_logical_adapter_number;

  a->vscapi_mode = vscapi_mode;

  diva_capi_read_cfg_value (&sync_req,
                            card->Id,
                            "XDI\\AN\\GroupOpt",
                            DivaCfgLibValueTypeUnsigned,
                            &a->group_optimization_enabled,
                            sizeof(a->group_optimization_enabled));
  diva_capi_read_cfg_value (&sync_req,
                            card->Id,
                            "XDI\\AN\\DynL1",
                            DivaCfgLibValueTypeBool,
                            &a->flag_dynamic_l1_down,
                            sizeof(a->flag_dynamic_l1_down));

  a->request = DIRequest; /* card->d.request; */
  a->max_listen = (a->vscapi_mode == 0) ? (card->PRI ? 8 : 2) : 2;
  k = card->d.channels + ((a->vscapi_mode == 0) ? 32 : 4) + a->max_listen;
#if IMPLEMENT_NULL_PLCI
  if (a->vscapi_mode == 0) {
    k += 2 + card->d.channels;
  }
#endif /* IMPLEMENT_NULL_PLCI */
  a->max_plci = (byte)((k > 255) ? 255 : k);
  if (!(a->plci = (PLCI *) vmalloc(sizeof(PLCI) * a->max_plci))) {
    printk(KERN_ERR "%s: failed alloc plci struct.\n", DRIVERLNAME);
		if (a->mtpx_cfg.cfg != 0)
			diva_os_free (0, a->mtpx_cfg.cfg);
    memset(a, 0, sizeof(DIVA_CAPI_ADAPTER));
    return(0);
  }
  memset(a->plci, 0, sizeof(PLCI) * a->max_plci);

  for (k = 0; k < a->max_plci; k++)
    {
      a->Id = (byte)card->Id;
      a->plci[k].Sig.callback = sync_callback;
      a->plci[k].Sig.XNum = 1;
      a->plci[k].Sig.X = a->plci[k].XData;
      a->plci[k].Sig.user[0] = (word) (card->Id - 1);
      a->plci[k].Sig.user[1] = (word) k;
      a->plci[k].NL.callback = sync_callback;
      a->plci[k].NL.XNum = 1;
      a->plci[k].NL.X = a->plci[k].XData;
      a->plci[k].NL.user[0] = (word) ((card->Id - 1) | 0x8000);
      a->plci[k].NL.user[1] = (word) k;
      a->plci[k].adapter = a;
    }

  a->profile.Number = max_adapter;
  a->profile.Channels = card->d.channels;
  if (card->d.features & DI_FAX3)
    {
      a->profile.Global_Options = 0x71;
      if (card->d.features & DI_CODEC)
        a->profile.Global_Options |= 0x6;
#if IMPLEMENT_DTMF
      a->profile.Global_Options |= 0x8;
#endif /* IMPLEMENT_DTMF */
#if (IMPLEMENT_LINE_INTERCONNECT || IMPLEMENT_LINE_INTERCONNECT2)
      if (a->vscapi_mode == 0) {
        a->profile.Global_Options |= 0x80;
      }
#endif /* (IMPLEMENT_LINE_INTERCONNECT || IMPLEMENT_LINE_INTERCONNECT2) */
#if IMPLEMENT_ECHO_CANCELLER
      a->profile.Global_Options |= 0x300;
#endif /* IMPLEMENT_ECHO_CANCELLER */
      a->profile.B1_Protocols = 0x3df;
      a->profile.B2_Protocols = 0x1fdf;
      a->profile.B3_Protocols = 0xbf;
      a->manufacturer_features = MANUFACTURER_FEATURE_HARDDTMF;
    }
  else
    {
      a->profile.Global_Options = 0x71;
      if (card->d.features & DI_CODEC)
        a->profile.Global_Options |= 0x2;
      a->profile.B1_Protocols = 0x43;
      a->profile.B2_Protocols = 0x1f0f;
      a->profile.B3_Protocols = 0x0f;
      a->manufacturer_features = 0;
    }

  if (a->vscapi_mode == 0) {
#if IMPLEMENT_LINE_INTERCONNECT
  a->li_pri = card->PRI;
  if (a->li_pri)
    {
      if (!(a->li_config.pri = (LI_CONFIG_PRI *) kmalloc(sizeof(LI_CONFIG_PRI), GFP_KERNEL))) {
        printk(KERN_ERR "%s: failed alloc li_config.pri struct.\n", DRIVERLNAME);
         if (a->mtpx_cfg.cfg != 0)
           diva_os_free (0, a->mtpx_cfg.cfg);
        memset(a, 0, sizeof(DIVA_CAPI_ADAPTER));
        return(0);
      }
      memset(a->li_config.pri, 0, sizeof(LI_CONFIG_PRI));
    }
  else
    {
      if (!(a->li_config.bri = (LI_CONFIG_BRI *) kmalloc(sizeof(LI_CONFIG_BRI), GFP_KERNEL))) {
        printk(KERN_ERR "%s: failed alloc li_config.bri struct.\n", DRIVERLNAME);
        if (a->mtpx_cfg.cfg != 0)
          diva_os_free (0, a->mtpx_cfg.cfg);
        memset(a, 0, sizeof(DIVA_CAPI_ADAPTER));
        return(0);
      }
      memset(a->li_config.bri, 0, sizeof(LI_CONFIG_BRI));
    }
#endif /* IMPLEMENT_LINE_INTERCONNECT */
#if IMPLEMENT_LINE_INTERCONNECT2
  a->li_pri = card->PRI;
  a->li_channels = a->li_pri ?
#if IMPLEMENT_NULL_PLCI
    MIXER_TOTAL_CHANNELS_PRI : MIXER_TOTAL_CHANNELS_BRI;
#else /* IMPLEMENT_NULL_PLCI */
    MIXER_CHANNELS_PRI : MIXER_CHANNELS_BRI;
#endif /* IMPLEMENT_NULL_PLCI */
  a->li_base = 0;
  for (i = 0; &adapter[i] != a; i++)
  {
    if (adapter[i].request)
      a->li_base = adapter[i].li_base + adapter[i].li_channels;
  }
  k = li_total_channels + a->li_channels;
  table_size = ((k * sizeof(LI_CONFIG) + 3) & ~3) + (2 * k) * ((k + 3) & ~3)+4;
  if (table_size>131072) {
    new_li_config_table = (LI_CONFIG *) vmalloc(table_size);
    if (new_li_config_table == NULL)
    {
      printk(KERN_ERR "%s: failed vmalloc li_config table.\n", DRIVERLNAME);
      if (a->mtpx_cfg.cfg != 0)
        diva_os_free (0, a->mtpx_cfg.cfg);
      memset(a, 0, sizeof(DIVA_CAPI_ADAPTER));
      return(0);
    }
    ((int *) new_li_config_table)[0] = 1;
  } else {
    new_li_config_table = (LI_CONFIG *) kmalloc(table_size, GFP_KERNEL);
    if (new_li_config_table == NULL)
    {
      printk(KERN_ERR "%s: failed kmalloc li_config table.\n", DRIVERLNAME);
      if (a->mtpx_cfg.cfg != 0)
        diva_os_free (0, a->mtpx_cfg.cfg);
      memset(a, 0, sizeof(DIVA_CAPI_ADAPTER));
      return(0);
    }
    ((int *) new_li_config_table)[0] = 0;
  }
  cp = (byte *) new_li_config_table;
  cp += 4;
  new_li_config_table = (LI_CONFIG *)cp;
  j = 0;
  for (i = 0; i < k; i++)
  {
    if ((i >= a->li_base) && (i < a->li_base + a->li_channels))
      memset(&new_li_config_table[i], 0, sizeof(LI_CONFIG));
    else
      memcpy(&new_li_config_table[i], &li_config_table[j], sizeof(LI_CONFIG));
    new_li_config_table[i].flag_table = ((byte *) new_li_config_table) +
      (((k * sizeof(LI_CONFIG) + 3) & ~3) + (2 * i) * ((k + 3) & ~3));
    new_li_config_table[i].coef_table = ((byte *) new_li_config_table) +
      (((k * sizeof(LI_CONFIG) + 3) & ~3) + (2 * i + 1) * ((k + 3) & ~3));
    if ((i >= a->li_base) && (i < a->li_base + a->li_channels))
    {
      new_li_config_table[i].adapter = a;
      memset(&new_li_config_table[i].flag_table[0], 0, k);
      memset(&new_li_config_table[i].coef_table[0], 0, k);
    }
    else
    {
      if (a->li_base != 0)
      {
        memcpy(&new_li_config_table[i].flag_table[0],
               &li_config_table[j].flag_table[0],
               a->li_base);
        memcpy(&new_li_config_table[i].coef_table[0],
               &li_config_table[j].coef_table[0],
               a->li_base);
      }
      memset(&new_li_config_table[i].flag_table[a->li_base], 0, a->li_channels);
      memset(&new_li_config_table[i].coef_table[a->li_base], 0, a->li_channels);
      if (a->li_base + a->li_channels < k)
      {
        memcpy(&new_li_config_table[i].flag_table[a->li_base + a->li_channels],
               &li_config_table[j].flag_table[a->li_base],
               k - (a->li_base + a->li_channels));
        memcpy(&new_li_config_table[i].coef_table[a->li_base + a->li_channels],
               &li_config_table[j].coef_table[a->li_base],
               k - (a->li_base + a->li_channels));
      }
      j++;
    }
  }
  li_total_channels = k;
  if (li_config_table != NULL) {
    cp = (byte *) li_config_table;
    cp -= 4;
    li_config_table = (LI_CONFIG *)cp;
    i = ((int *) li_config_table)[0];
    if (i==0)
      kfree(li_config_table);
    else
      vfree(li_config_table);
  }
  li_config_table = new_li_config_table;
  for (i = card->Id; i < max_adapter; i++)
  {
    if (adapter[i].request)
      adapter[i].li_base += a->li_channels;
  }
#endif /* IMPLEMENT_LINE_INTERCONNECT2 */
#if IMPLEMENT_NULL_PLCI
  if (card->PRI)
  {
    a->null_plci_base = 32;
    a->null_plci_channels = 32;
  }
  else
  {
    a->null_plci_base = 5;
    a->null_plci_channels = 4;
  }
#endif /* IMPLEMENT_NULL_PLCI */

  } else {
    a->li_channels = 0;
    a->li_base = 0;
    a->null_plci_channels = 0;
    a->null_plci_base = 0;
  }

  if (a == &adapter[max_adapter])
    max_adapter++;

	if (event) {
		spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);
		event->next = events;
		events      = event;
		spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
	}

  spin_lock(&ll_lock);
  card->next = cards;
  cards = card;
  spin_unlock(&ll_lock);

  AutomaticLaw(a);
  while(i++ < 30) {
    if (a->automatic_law > 3)
      break;
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout(HZ / 10 + 1);
  }

  /* profile information */
  ctrl->profile.nbchannel = card->d.channels;
  ctrl->profile.goptions = a->profile.Global_Options;
  ctrl->profile.support1 = a->profile.B1_Protocols;
  ctrl->profile.support2 = a->profile.B2_Protocols;
  ctrl->profile.support3 = a->profile.B3_Protocols;
  /* manufacturer profile information */
  ctrl->profile.manu[0] = a->profile.Manuf.private_options;
  ctrl->profile.manu[1] = a->profile.Manuf.reserved_1;
  ctrl->profile.manu[2] = a->profile.Manuf.reserved_2;
  ctrl->profile.manu[3] = a->profile.Manuf.reserved_3;
  ctrl->profile.manu[4] = a->profile.Manuf.reserved_4;

  ctrl->ready(ctrl);

  DBG_TRC(("adapter added, max_adapter=%d", max_adapter));
  return(1);
}

static void
clean_adapter(int id)
{
  DIVA_CAPI_ADAPTER *a;
#if IMPLEMENT_LINE_INTERCONNECT2
  int i, k;
  byte *cp;
#endif /* IMPLEMENT_LINE_INTERCONNECT2 */

  a = &adapter[id];
#if IMPLEMENT_LINE_INTERCONNECT
  if (a->li_pri)
    {
      if(a->li_config.pri)
        kfree(a->li_config.pri);
    }
  else
    {
      if(a->li_config.bri)
        kfree(a->li_config.bri);
    }
#endif /* IMPLEMENT_LINE_INTERCONNECT */
#if IMPLEMENT_LINE_INTERCONNECT2
  k = li_total_channels - a->li_channels;
  if (k == 0)
  {
    cp = (byte *) li_config_table;
    cp -= 4;
    li_config_table = (LI_CONFIG *)cp;
    i = ((int *) li_config_table)[0];
    if (i==0)
      kfree(li_config_table);
    else
      vfree(li_config_table);
    li_config_table = NULL;
  }
  else
  {
    if (a->li_base < k)
    {
      memmove(&li_config_table[a->li_base],
              &li_config_table[a->li_base + a->li_channels],
              (k - a->li_base) * sizeof(LI_CONFIG));
      for (i = 0; i < k; i++)
      {
        memmove(&li_config_table[i].flag_table[a->li_base],
                &li_config_table[i].flag_table[a->li_base + a->li_channels],
                k - a->li_base);
        memmove(&li_config_table[i].coef_table[a->li_base],
                &li_config_table[i].coef_table[a->li_base + a->li_channels],
                k - a->li_base);
      }
    }
  }
  li_total_channels = k;
  for (i = id; i < max_adapter; i++)
  {
    if (adapter[i].request)
      adapter[i].li_base -= a->li_channels;
  }
#endif /* IMPLEMENT_LINE_INTERCONNECT2 */
  if(a->plci)
    vfree(a->plci);

  if (a->mtpx_cfg.cfg != 0)
    diva_os_free (0, a->mtpx_cfg.cfg);

  memset(a, 0x00, sizeof(DIVA_CAPI_ADAPTER));
  while ((max_adapter != 0) && !adapter[max_adapter - 1].request)
    max_adapter--;
}

static void
didd_callback(void *context, DESCRIPTOR* adapter, int removal)
{
  if (adapter->type == IDI_DADAPTER)
  {
    printk(KERN_ERR "%s: Change in DAdapter ? Oops ?.\n", DRIVERLNAME);
    DBG_ERR(("Notification about IDI_DADAPTER change ! Oops."))
    return;
  }
  else if (adapter->type == IDI_MADAPTER)
  {
    if (diva_capi_uses_madapter && !removal) {
      DBG_LOG(("Ignore MADAPTER"))
      /* diva_add_card(adapter); */
    } else {
      printk(KERN_ERR "%s: Change in MAdapter ? Oops ?.\n", DRIVERLNAME);
      DBG_ERR(("Notification about IDI_MADAPTER change ! Oops."))
    }
    return;
  }
  else if (adapter->type == IDI_DIMAINT)
  {
    if (removal)
    {
      DbgDeregister();
      memset(&MAdapter, 0, sizeof(MAdapter));
      dprintf = no_printf;
    }
    else
    {
      memcpy(&MAdapter, adapter, sizeof(MAdapter));
      dprintf = (DIVA_DI_PRINTF)MAdapter.request;
      DbgRegister("CAPI20", DRIVERRELEASE, DBG_DEFAULT);
    }
  }
  else if ((adapter->type > 0) &&
           (adapter->type < 16))
  {  /* IDI Adapter */
    if (!diva_capi_uses_madapter) {
      if (removal)
      {
        diva_remove_card(adapter);
      }
      else
      {
        diva_add_card(adapter);
      }
    }
  }
}

static int
diva_os_dpc_proc(void* context)
{
  int i;
  diva_os_thread_dpc_t* psoft_isr = (diva_os_thread_dpc_t*)context;
  diva_os_soft_isr_t* pisr = psoft_isr->psoft_isr;

  atomic_inc(&psoft_isr->thread_running);
  if (atomic_read(&psoft_isr->thread_running) > 1) {
      printk(KERN_WARNING"%s: thread already running\n", DRIVERLNAME);
      return(0);
  }

  printk(KERN_INFO "%s: thread started with pid %d\n", DRIVERLNAME, current->pid);

  exit_mm(current);
  for (i = 0; (current->files && (i < current->files->max_fds)); i++ ) {
    if (current->files->fd[i]) close(i);
  }

  strcpy(current->comm, pisr->dpc_thread_name);

  for(;;) {
    down_interruptible(&psoft_isr->divas_sem);
    if(!(atomic_read(&psoft_isr->thread_running)))
      break;
    if(signal_pending(current)) {
      /* we may want to do something on signals here */
      spin_lock_irq(&current->sigmask_lock);
      flush_signals(current);
      spin_unlock_irq(&current->sigmask_lock);
    } else {
     (*(pisr->callback))(pisr, pisr->callback_context);
    }
  }

  up(&psoft_isr->divas_end);
  psoft_isr->pid = -1;
  return (0);
}

int
diva_os_initialize_soft_isr (diva_os_soft_isr_t* psoft_isr, diva_os_soft_isr_callback_t callback, void*   callback_context)
{
  diva_os_thread_dpc_t* context;
  int name_len;

  context = (diva_os_thread_dpc_t*)kmalloc(sizeof(*context), GFP_KERNEL);

  name_len = MIN(sizeof(current->comm), sizeof(psoft_isr->dpc_thread_name));
  psoft_isr->dpc_thread_name[name_len - 1] = 0;

  psoft_isr->object = context;
  if (!psoft_isr->object) {
    return (-1);
  }
  memset (context, 0x00, sizeof (*context));
  psoft_isr->callback = callback;
  psoft_isr->callback_context = callback_context;
  context->psoft_isr = psoft_isr;
  init_MUTEX_LOCKED(&context->divas_sem);
  init_MUTEX_LOCKED(&context->divas_end);
  context->pid       = -1;

  context->pid = kernel_thread (diva_os_dpc_proc, context, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
  return ((context->pid == (-1)) ? (-1) : 0);
}

int
diva_os_schedule_soft_isr (diva_os_soft_isr_t* psoft_isr)
{
  diva_os_thread_dpc_t* context = (diva_os_thread_dpc_t*)psoft_isr->object;

  up(&context->divas_sem);
  return (1);
}

void
diva_os_remove_soft_isr (diva_os_soft_isr_t* psoft_isr)
{
  diva_os_thread_dpc_t* context;

  if (psoft_isr->object) {
    context = (diva_os_thread_dpc_t*)psoft_isr->object;
    if (context->pid != -1) {
      atomic_set(&context->thread_running, 0);
      up(&context->divas_sem);
      down_interruptible(&context->divas_end);
    }
    kfree(psoft_isr->object);
    psoft_isr->object = 0;
  }
}

void diva_os_add_adapter_descriptor (DESCRIPTOR* d, int operation) {
  IDI_SYNC_REQ req;

  if (operation) { /* Add    */
    req.didd_remove_adapter.e.Req = 0;
    req.didd_add_adapter.e.Rc = IDI_SYNC_REQ_DIDD_ADD_ADAPTER;
    req.didd_add_adapter.info.descriptor = (void *)d;
    DAdapter.request((ENTITY *)&req);
  } else {         /* Remove */
    req.didd_remove_adapter.e.Req = 0;
    req.didd_remove_adapter.e.Rc = IDI_SYNC_REQ_DIDD_REMOVE_ADAPTER;
    req.didd_remove_adapter.info.p_request = d->request;
    DAdapter.request((ENTITY *)&req);
  }
}

void
diva_os_sleep (dword mSec)
{
  unsigned long timeout = HZ * mSec / 1000 + 1;

  set_current_state(TASK_UNINTERRUPTIBLE);
  schedule_timeout(timeout);
}

void* diva_os_malloc (unsigned long flags, unsigned long size)
{
  void* ret = NULL;

  if (size) {
      ret = (void*)vmalloc ((unsigned int)size);
  }
  return (ret);
}

void
diva_os_free (unsigned long unused, void* ptr)
{
  if (ptr) {
    vfree (ptr);
  }
}



/* -------------------------------------------------------
    Spin Lock framework
   ------------------------------------------------------- */
void diva_os_enter_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  spin_lock(a);
}

void diva_os_leave_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  spin_unlock(a);
}


void diva_os_api_lock   (diva_os_spin_lock_magic_t* old_irql) {
  spin_lock(&api_lock);
  spin_lock(&ll_lock);
}

void diva_os_api_unlock (diva_os_spin_lock_magic_t* old_irql) {
  spin_unlock(&ll_lock);
  spin_unlock(&api_lock);
}

const DIVA_CAPI_ADAPTER* diva_os_get_adapter (int nr) {
  diva_card *p = cards;

  while (p && nr--) {
    p = p->next;
  }

  return (p ? p->adapter : 0);
}

static int __init
connect_didd(void)
{
  int x = 0;
  IDI_SYNC_REQ req;
  DESCRIPTOR DIDD_Table[MAX_DESCRIPTORS];

  DIVA_DIDD_Read(DIDD_Table, sizeof(DIDD_Table));

  /*
    Locate DADAPTER
    */
	for (x = 0; x < MAX_DESCRIPTORS; x++) {
		if (DIDD_Table[x].type == IDI_DADAPTER) {	/* DADAPTER found */
			memcpy(&DAdapter, &DIDD_Table[x], sizeof(DAdapter));
			break;
		}
	}

	if (DAdapter.request == 0) {
		return (0);
	}

  /*
    First connect to the MAINT
    */
  for (x = 0; x < MAX_DESCRIPTORS; x++)
  {
    if (DIDD_Table[x].type == IDI_DIMAINT)
    {
      memcpy(&MAdapter, &DIDD_Table[x], sizeof(DAdapter));
      dprintf = (DIVA_DI_PRINTF)MAdapter.request;
      DbgRegister("CAPI20", DRIVERRELEASE, DBG_DEFAULT);
    }
  }

  /*
	  Start capi event queue if necessary
		*/
  diva_init_event_queue();

  for (x = 0; x < MAX_DESCRIPTORS; x++)
  {
    if (DIDD_Table[x].type == IDI_MADAPTER)
    { /* MADAPTER found */
      diva_capi_uses_madapter = 1;
      DBG_LOG(("Add MADAPTER"))
      diva_add_card(&DIDD_Table[x]);
    }
  }

  /*
    Look for IDI adapters only if MADAPTER was not found
    */
  if (!diva_capi_uses_madapter) {
    for (x = 0; x < MAX_DESCRIPTORS; x++) {
      if (DIDD_Table[x].type && (DIDD_Table[x].type < 16) && (DIDD_Table[x].channels))
      {  /* IDI Adapter found */
        diva_add_card(&DIDD_Table[x]);
      }
    }
  }

  /*
     Register with DIDD
    */
  req.didd_notify.e.Req = 0;
  req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REGISTER_ADAPTER_NOTIFY;
  req.didd_notify.info.callback = didd_callback;
  req.didd_notify.info.context = 0;
  DAdapter.request((ENTITY *)&req);
  if (req.didd_notify.e.Rc != 0xff)
    return(0);
  notify_handle = req.didd_notify.info.handle;

  return(1);
}

static int __init
init_main_structs(void)
{
  if(!(mapped_msg = (CAPI_MSG *) kmalloc(MAX_MSG_SIZE, GFP_KERNEL))) {
    printk(KERN_ERR "%s: failed alloc mapped_msg.\n", DRIVERLNAME);
    return 0;
  }

  memset(ControllerMap, 0, MAX_ADAPTER + 1);

  /* TODO: maybe we can alloc this piece by piece when needed */
  if (!(adapter = vmalloc(sizeof(DIVA_CAPI_ADAPTER) * MAX_ADAPTER))) {
    printk(KERN_ERR "%s: failed alloc adapter struct.\n", DRIVERLNAME);
    kfree(mapped_msg);
    return 0;
  }
  memset(adapter, 0, sizeof(DIVA_CAPI_ADAPTER) * MAX_ADAPTER);

  /* TODO: maybe we can alloc this piece by piece when needed */
  if (!(application = vmalloc(sizeof(APPL) * MAX_APPL))) {
    printk(KERN_ERR "%s: failed alloc application struct.\n", DRIVERLNAME);
    kfree(mapped_msg);
    vfree(adapter);
    return 0;
  }
  memset(application, 0, sizeof(APPL) * MAX_APPL);

#ifdef C_IND_RESOURCE_INDEX_ENTRIES
  if (!(c_ind_resource_index = vmalloc(sizeof(struct resource_index_entry) * C_IND_RESOURCE_INDEX_ENTRIES))) {
    printk(KERN_ERR "%s: failed alloc resource index struct.\n", DRIVERLNAME);
    kfree(mapped_msg);
    vfree(adapter);
    vfree(application);
    return 0;
  }
  memset(c_ind_resource_index, 0, sizeof(struct resource_index_entry) * C_IND_RESOURCE_INDEX_ENTRIES);
#endif

  return(1);
}

static void __exit
remove_main_structs(void)
{
  if (c_ind_resource_index)
    vfree(c_ind_resource_index);
  if (application)
    vfree(application);
  if (adapter)
    vfree(adapter);
  if (mapped_msg)
    kfree(mapped_msg);
}

static void __exit
disconnect_didd(void)
{
  IDI_SYNC_REQ req;

  req.didd_notify.e.Req = 0;
  req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REMOVE_ADAPTER_NOTIFY;
  req.didd_notify.info.handle = notify_handle;
  DAdapter.request((ENTITY *)&req);
}

static int __init
divacapi_init(void)
{
  char tmprev[32];
  int ret = 0;

  MOD_INC_USE_COUNT;

  api_lock = SPIN_LOCK_UNLOCKED;
  ll_lock = SPIN_LOCK_UNLOCKED;
  mem_q_lock = SPIN_LOCK_UNLOCKED;

  diva_q_init (&mem_release_q);

  sprintf(DRIVERRELEASE, "%s", DRRELEXTRA);

  printk(KERN_INFO "%s\n", DRIVERNAME);
  printk(KERN_INFO "%s: Rel:%s  Rev:", DRIVERLNAME, DRIVERRELEASE);
  strcpy(tmprev, main_revision);
  printk("%s  Build: %s(%s)\n", getrev(tmprev),
        diva_capi_common_code_build, DIVA_BUILD);

  strcpy(divas_driver.name, DRIVERLNAME);
  strcpy(tmprev, main_revision);
  sprintf(divas_driver.revision, "%s/%s %s(%s)", DRIVERRELEASE,
          getrev(tmprev), diva_capi_common_code_build, DIVA_BUILD);

  di = attach_capi_driver(&divas_driver);

  if (!di) {
		diva_shutdown_event_queue();
    printk(KERN_ERR "%s: failed to attach capi_driver.\n", DRIVERLNAME);
    ret = -EIO;
    goto out;
  }

  if(!init_main_structs()) {
    printk(KERN_ERR "%s: failed to init main structs.\n", DRIVERLNAME);
    detach_capi_driver(&divas_driver);
		diva_shutdown_event_queue();
    ret = -EIO;
    goto out;
  }

  if(!connect_didd()) {
    int count = 100;
    word i = 1;

    printk(KERN_ERR "%s: failed to connect to DIDD.\n", DRIVERLNAME);

    while(i && count--)
    {
      spin_lock(&api_lock);
      i = api_remove_start();
      spin_unlock(&api_lock);
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout(10);
    }

    if (i)
      printk(KERN_WARNING "%s: could not remove signaling ID's.\n", DRIVERLNAME);

    diva_remove_cards();

    detach_capi_driver(&divas_driver);
    remove_main_structs();
		diva_shutdown_event_queue();

    ret = -EIO;

    goto out;
  }

	diva_start_management (1000);

out:
  MOD_DEC_USE_COUNT;
  return ret;
}

static int diva_capi_read_cfg_value (IDI_SYNC_REQ* sync_req,
                                     int instance,
                                     const char* name,
                                     diva_cfg_lib_value_type_t value_type,
                                     void* data,
                                     dword data_length)
{
  diva_cfg_lib_return_type_t ret = DivaCfgLibNotFound;

  memset (sync_req, 0x00, sizeof(*sync_req));
  sync_req->didd_get_cfg_lib_ifc.e.Rc = IDI_SYNC_REQ_DIDD_GET_CFG_LIB_IFC;
  DAdapter.request((ENTITY *)sync_req);

  if (sync_req->didd_get_cfg_lib_ifc.e.Rc == 0xff)
  {
    diva_cfg_lib_access_ifc_t* cfg_lib_ifc = sync_req->didd_get_cfg_lib_ifc.info.ifc;

    if (cfg_lib_ifc != 0)
    {
      ret = cfg_lib_ifc->diva_cfg_storage_read_cfg_var_proc(cfg_lib_ifc,
                           TargetCapiDrv, instance, 0, 0, name,
                           value_type, &data_length, data);

      (*(cfg_lib_ifc->diva_cfg_lib_free_cfg_interface_proc))(cfg_lib_ifc);

      if (ret == DivaCfgLibOK) {
        DBG_LOG(("CAPI(%d) read %s configuration value", instance, name))
      } else {
        DBG_ERR(("CAPI(%d) failed to read %s configuration value", instance, name))
      }
    }
  }

  return ((ret == DivaCfgLibOK) ? 0 : -1);
}

static void diva_process_mem_release_q(int force)
{
  diva_mem_release_t* mem_release;
  diva_entity_queue_t q;

  diva_q_init (&q);

  spin_lock(&mem_q_lock);

  mem_release = (diva_mem_release_t*)diva_q_get_head(&mem_release_q);

  while (mem_release != 0)
  {
    if (force != 0 || time_after(jiffies, mem_release->t + 2*HZ))
    {
      diva_mem_release_t* next = (diva_mem_release_t*)diva_q_get_next(&mem_release->link);
      diva_q_remove(&mem_release_q, &mem_release->link);
      diva_q_add_tail(&q, &mem_release->link);
      mem_release = next;
    }
    else
    {
      mem_release = (diva_mem_release_t*)diva_q_get_next(&mem_release->link);
    }
  }

  spin_unlock(&mem_q_lock);

  while ((mem_release = (diva_mem_release_t*)diva_q_get_head(&q)) != 0)
  {
    diva_q_remove (&q, &mem_release->link);

    if (mem_release->DataNCCI)
      kfree(mem_release->DataNCCI);
    if (mem_release->DataFlags)
      kfree(mem_release->DataFlags);
    if (mem_release->ReceiveBuffer)
      vfree(mem_release->ReceiveBuffer);
    if (mem_release->xbuffer_ptr) {
      int i;

      for (i = 0; i < mem_release->num_data; i++)
        if (mem_release->xbuffer_ptr[i]) kfree(mem_release->xbuffer_ptr[i]);
      kfree(mem_release->xbuffer_ptr);
    }
    if (mem_release->xbuffer_internal)
      kfree(mem_release->xbuffer_internal);
    if (mem_release->xbuffer_table)
      kfree(mem_release->xbuffer_table);

    kfree (mem_release);
  }
}

static void diva_capi_event_queue_proc (struct _diva_os_soft_isr* psoft_isr, void* Context) {
	diva_card_event_q_t* event;
	struct sk_buff *skb;
	unsigned long flags;

	spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);

	event = events;
	while (event != 0) {
		if (skb_queue_empty(&event->inqueue)) {
			event = event->next;
		} else {
			if ((skb = skb_dequeue(&event->inqueue)) != 0) {
				spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);

				diva_internal_send_message(event->card->capi_ctrl, skb);

				spin_lock_irqsave(&diva_capi_event_queue.event_q_lock, flags);
			}
			event = events; /* restart event processing */
		}
	}

	spin_unlock_irqrestore(&diva_capi_event_queue.event_q_lock, flags);
}

static void diva_init_event_queue (void) {
	IDI_SYNC_REQ sync_req;

	diva_capi_read_cfg_value (&sync_req,
                            0,
														"SendMessageInInterruptContext",
                            DivaCfgLibValueTypeBool,
                            &diva_handle_send_message_in_interrupt_context,
                            sizeof(diva_handle_send_message_in_interrupt_context));

	if (diva_handle_send_message_in_interrupt_context != 0) {
		strcpy (diva_capi_event_queue.queue_soft_isr.dpc_thread_name, "kcapievd");
		if (diva_os_initialize_soft_isr (&diva_capi_event_queue.queue_soft_isr,
		                                 diva_capi_event_queue_proc, &diva_capi_event_queue)) {
			diva_handle_send_message_in_interrupt_context = 0;
			DBG_ERR(("Failed to initialize event queue"))
		} else {
			diva_capi_event_queue.event_q_lock = SPIN_LOCK_UNLOCKED;
		}
	}
}

static void diva_shutdown_event_queue(void) {
	if (diva_handle_send_message_in_interrupt_context != 0) {
		diva_os_remove_soft_isr (&diva_capi_event_queue.queue_soft_isr);
	}
}

static void __exit
divacapi_exit(void)
{
  int count = 100;
  word ret = 1;

  while(ret && count--)
  {
    spin_lock(&api_lock);
    ret = api_remove_start();
    spin_unlock(&api_lock);
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout(10);
  }

  if (ret)
    printk(KERN_WARNING "%s: could not remove signaling ID's.\n", DRIVERLNAME);

  diva_stop_management ();
  disconnect_didd();
  diva_remove_cards();
  detach_capi_driver(&divas_driver);
  remove_main_structs();
	diva_shutdown_event_queue();
  diva_process_mem_release_q(1);
  DbgDeregister();
  printk(KERN_INFO "%s: module unloaded.\n", DRIVERLNAME);
}

module_init(divacapi_init);
module_exit(divacapi_exit);
