/* $Id: divamnt.c,v 1.1.2.4 2001/05/01 15:48:05 armin Exp $
 *
 * Driver for Eicon DIVA Server ISDN cards.
 * Maint module
 *
 * Copyright 2000-2007 by Armin Schindler (armin@melware.de)
 * Copyright 2000-2007 Cytronics & Melware (info@melware.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY OF ANY KIND WHATSOEVER INCLUDING ANY 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define FEATURE_NOTIFIER_SUPPORTED 0

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
#if FEATURE_NOTIFIER_SUPPORTED
#include <linux/notifier.h>
#include <linux/reboot.h>
#endif
#include <linux/console.h>
#endif

#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/vmalloc.h>


#include "platform.h"
#include "di_defs.h"
#include "divasync.h"
#include "di_defs.h"
#include "debug_if.h"
#include "um_dbg.h"

EXPORT_NO_SYMBOLS;

static char *main_revision = "$Revision: 1.1.2.4 $";

MODULE_DESCRIPTION(             "Maint driver for Eicon DIVA Server cards");
MODULE_AUTHOR(                  "Cytronics & Melware, Eicon Networks");
MODULE_SUPPORTED_DEVICE(        "DIVA card driver");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

long buffer_length = 128;
MODULE_PARM(buffer_length, "l");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
MODULE_PARM_DESC(buffer_length, "Trace buffer length in kBytes");
#endif
long diva_dbg_mem=0;
MODULE_PARM(diva_dbg_mem, "l");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
MODULE_PARM_DESC(diva_dbg_mem, "External device memory address");
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
int diva_register_notifier = 0;
MODULE_PARM(diva_register_notifier, "i");
MODULE_PARM_DESC(diva_register_notifier, "Register system notifier (0,1)");
#endif

static char *DRIVERNAME = "Eicon DIVA - MAINT module (http://www.melware.net)";
static char *DRIVERLNAME = "diva_mnt";
static char *DRIVERRELEASE = "3.0.10-107.884-1";

#define DBG_MINIMUM  (DL_LOG + DL_FTL + DL_ERR)
#define DBG_DEFAULT  (DBG_MINIMUM + DL_XLOG + DL_REG)

extern void DIVA_DIDD_Read (void *, int);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
extern void maint_write_system_info (unsigned long event, const char* info, int length);
static int diva_notifier_registered;
static int diva_reboot_notifier_registered;
#endif

static dword notify_handle;
static DESCRIPTOR DAdapter;
static DESCRIPTOR MAdapter;
static DESCRIPTOR MaintDescriptor = { IDI_DIMAINT, 0, 0, (IDI_CALL)diva_maint_prtComp };

static wait_queue_head_t msgwaitq;
static atomic_t opened; 
static struct timeval start_time;


void no_printf (unsigned char * x ,...)
{
  /* dummy debug function */
}
DIVA_DI_PRINTF dprintf = no_printf;


#include "debuglib.c"

/*
**  helper functions
*/
static char *
getrev(const char *revision)
{
  char *rev;
  char *p;

  if ((p = strchr(revision, ':'))) {
    rev = p + 2;
    p = strchr(rev, '$');
    *--p = 0;
  } else rev = "1.0";

  return rev;
} 

void diva_os_enter_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  unsigned long flags;

  spin_lock_irqsave (a, flags);
  *old_irql = (diva_os_spin_lock_magic_t)flags;
}

void diva_os_leave_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  unsigned long flags = (unsigned long)*old_irql;

	spin_unlock_irqrestore (a, flags);
}

void diva_os_get_time (dword* sec, dword* usec) {
  struct timeval tv;

  do_gettimeofday (&tv);

  if (tv.tv_sec > start_time.tv_sec) {
    if (start_time.tv_usec > tv.tv_usec) {
      tv.tv_sec--;
      tv.tv_usec += 1000000;
    }
    *sec  = (dword)(tv.tv_sec  - start_time.tv_sec);
    *usec = (dword)(tv.tv_usec - start_time.tv_usec);
  } else if (tv.tv_sec == start_time.tv_sec) {
    *sec  = 0;
    if (start_time.tv_usec < tv.tv_usec) {
      *usec = (dword)(tv.tv_usec - start_time.tv_usec);
    } else {
      *usec = 0;
    }
  } else {
    *sec  = (dword)tv.tv_sec;
    *usec = (dword)tv.tv_usec;
  }
}

/*
** /proc entries
*/

extern struct proc_dir_entry *proc_net_isdn_eicon;
static struct proc_dir_entry *maint_proc_entry = NULL;
static struct proc_dir_entry *maint_dbg_proc_entry = NULL;

/*
	Read function is provided for compatibility reason - this allows
  to read unstructured traces, formated as ascii string only
  */
static ssize_t
maint_read(struct file *file, char *buf, size_t count, loff_t * off)
{
  diva_dbg_entry_head_t* pmsg = 0;
  diva_os_spin_lock_magic_t old_irql;
  word size;
  char* pstr, *dli_label = "UNK";
	int str_length;
	int *str_msg;

  if (off != &file->f_pos)
    return -ESPIPE;

  if (!file->private_data) {
    for (;;) {
      while ((pmsg = diva_maint_get_message (&size, &old_irql))) {
        if (!(pmsg->facility == MSG_TYPE_STRING)) {
          diva_maint_ack_message (1, &old_irql);
        } else {
          break;
        }
      }

      if (!pmsg) {
        if (file->f_flags & O_NONBLOCK) {
          return (-EAGAIN);
        }
        interruptible_sleep_on(&msgwaitq);
        if (signal_pending(current)) {
          return (-ERESTARTSYS);
        }
      } else {
        break;
      }
    }
    /*
      The length of message that shoule be read is:
      pmsg->data_length + label(25) + DrvID(2) + byte CR + trailing zero
      */
    if (!(str_msg = (int*)kmalloc(pmsg->data_length + 29 + 2*sizeof(int), GFP_ATOMIC))) {
      diva_maint_ack_message (0, &old_irql);
      return (-ENOMEM);
    }
    pstr = (char*)&str_msg[2];

    switch (pmsg->dli) {
      case DLI_LOG :
        dli_label = "LOG";
        break;
      case DLI_FTL :
        dli_label = "FTL";
        break;
      case DLI_ERR :
        dli_label = "ERR";
        break;
      case DLI_TRC :
        dli_label = "TRC";
        break;
      case DLI_REG :
        dli_label = "REG";
        break;
      case DLI_MEM :
        dli_label = "MEM";
        break;
      case DLI_SPL :
        dli_label = "SPL";
        break;
      case DLI_IRP :
        dli_label = "IRP";
        break;
      case DLI_TIM :
        dli_label = "TIM";
        break;
      case DLI_TAPI:
        dli_label = "TAPI";
        break;
      case DLI_NDIS:
        dli_label = "NDIS";
        break;
      case DLI_CONN:
        dli_label = "CONN";
        break;
      case DLI_STAT:
        dli_label = "STAT";
        break;
      case DLI_PRV0:
        dli_label = "PRV0";
        break;
      case DLI_PRV1:
        dli_label = "PRV1";
        break;
      case DLI_PRV2:
        dli_label = "PRV2";
        break;
      case DLI_PRV3:
        dli_label = "PRV3";
        break;
    }
    str_length = sprintf (pstr, "%s %02x %s\n",
                          dli_label, (byte)pmsg->drv_id, (char*)&pmsg[1]);
    str_msg[0] = str_length;
    str_msg[1] = 0;
    file->private_data = str_msg;
    diva_maint_ack_message (1, &old_irql);
  } else {
    str_msg = (int*)file->private_data;
    pstr = (char*)&str_msg[2];
    pstr += str_msg[1]; /* head + offset */
    str_length = str_msg[0] - str_msg[1]; /* length - offset */
  }
  str_length = MIN(str_length, count);

  if (copy_to_user (buf, pstr, str_length)) {
    kfree (str_msg);
    file->private_data = 0;
    return (-EFAULT);
  }
  str_msg[1] += str_length;
  if ((str_msg[0] - str_msg[1]) <= 0) {
    kfree (str_msg);
    file->private_data = 0;
  }

  return (str_length);
}

static ssize_t
maint_write(struct file *file, const char *buf, size_t count, loff_t * off)
{
  return (-ENODEV);
}                                  

static unsigned int
maint_poll(struct file *file, poll_table * wait)
{
  unsigned int mask = 0;

  poll_wait(file, &msgwaitq, wait);
  mask = POLLOUT | POLLWRNORM;
  if (file->private_data || diva_dbg_q_length ()) {
    mask |= POLLIN | POLLRDNORM;
  }
  return (mask);
}

static int
maint_open(struct inode *ino, struct file *filep)
{
  if (atomic_read(&opened))
      return(-EBUSY);
  else
    atomic_inc(&opened);

  filep->private_data = 0;

  return(0);
}

static int
maint_close(struct inode *ino, struct file *filep)
{

  if (filep->private_data) {
    kfree(filep->private_data);
    filep->private_data = 0;
  }

  atomic_dec(&opened);
  return(0);
}

static int maint_ioctl (struct inode *inode,
                        struct file *file,
                        uint io_cmd, ulong arg) {
  byte data[128];
  dword cmd, id, mask;
  int ret;

  if (copy_from_user((void *)&data[0], (void *)arg, 3*sizeof(dword))) {
    return(-EFAULT);
  }

  cmd  = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24); /* command */
  id   = data[4] | (data[5] << 8) | (data[6] << 16) | (data[7] << 24); /* driver id */
  mask = data[8] | (data[9] << 8) | (data[10] << 16) | (data[11] << 24); /* mask or size */

  switch (cmd) {
    case DITRACE_CMD_GET_DRIVER_INFO:
      if ((ret = diva_get_driver_info (id, data, sizeof(data))) > 0) {
        if (copy_to_user ((void*)arg, (void*)&data[0], ret))
          ret = -EFAULT;
      } else {
        ret = -EINVAL;
      }
      break;

    /*
       Filter commans will ignore the ID due to fact that filtering affects
       the B- channel and Audio Tap trace levels only. Also MAINT driver will
       select the right trace ID self
       */
    case DITRACE_WRITE_SELECTIVE_TRACE_FILTER:
      if (!mask) {
        ret = diva_set_trace_filter (1, "*");
      } else if (mask < sizeof(data)) {
        if (copy_from_user((void *)&data[0], (void *)(arg+12), mask)) {
          ret = -EFAULT;
        } else {
          ret = diva_set_trace_filter ((int)mask, data);
        }
      } else {
        ret = -EINVAL;
      }
      break;

    case DITRACE_READ_SELECTIVE_TRACE_FILTER:
      if ((ret = diva_get_trace_filter (sizeof(data), data)) > 0) {
        if (copy_to_user ((void*)arg, (void*)&data[0], ret))
          ret = -EFAULT;
      } else {
        ret = -ENODEV;
      }
      break;

    case DITRACE_READ_DRIVER_DBG_MASK:
      if ((ret = diva_get_driver_dbg_mask (id, (byte*)data)) > 0) {
        if (copy_to_user ((void*)arg, (void*)&data[0], ret))
          ret = -EFAULT;
      } else {
        ret = -ENODEV;
      }
      break;

    case DITRACE_WRITE_DRIVER_DBG_MASK:
      if ((ret = diva_set_driver_dbg_mask (id, mask)) <= 0) {
        ret = -ENODEV;
      }
      break;

    case DITRACE_READ_TRACE_ENTRY: {
      diva_os_spin_lock_magic_t old_irql;
      word size;
      diva_dbg_entry_head_t* pmsg;
      byte* pbuf;

      if ((pmsg = diva_maint_get_message (&size, &old_irql))) {
        if (size > mask) {
          diva_maint_ack_message (0, &old_irql);
          ret = -EINVAL;
        } else {
          if (!(pbuf = kmalloc (size, GFP_ATOMIC))) {
            diva_maint_ack_message (0, &old_irql);
            ret = -ENOMEM;
          } else {
            ret = size;
            memcpy (pbuf, pmsg, size);
            diva_maint_ack_message (1, &old_irql);
            if (copy_to_user ((void*)arg, (void*)pbuf, size))
              ret = -EFAULT;
            kfree (pbuf);
          }
        }
      } else {
        ret = 0;
      }
    } break;

    case DITRACE_READ_TRACE_ENTRYS: {
      diva_os_spin_lock_magic_t old_irql;
      word size;
      diva_dbg_entry_head_t* pmsg;
      byte* pbuf = 0;
      int written = 0;

      if (mask < 4096) {
        ret = -EINVAL;
        break;
      }
      if (!(pbuf = kmalloc (mask, GFP_KERNEL))) {
        return (-ENOMEM);
      }

      for (;;) {
        if (!(pmsg = diva_maint_get_message (&size, &old_irql))) {
          break;
        }
        if ((size+8) > mask) {
          diva_maint_ack_message (0, &old_irql);
          break;
        }
        /*
          Write entry length
          */
        pbuf[written++] = (byte)size;
        pbuf[written++] = (byte)(size >> 8);
        pbuf[written++] = 0;
        pbuf[written++] = 0;
        /*
          Write message
          */
        memcpy (&pbuf[written], pmsg, size);
        diva_maint_ack_message (1, &old_irql);
        written += size;
        mask -= (size+4);
      }
      pbuf[written++] = 0;
      pbuf[written++] = 0;
      pbuf[written++] = 0;
      pbuf[written++] = 0;

      if (copy_to_user ((void*)arg, (void*)pbuf, written)) {
        ret = -EFAULT;
      } else {
        ret = written;
      }
      kfree (pbuf);
    } break;

    default:
      ret = -EINVAL;
  }

  return (ret);
}


static loff_t
maint_lseek(struct file *file, loff_t offset, int orig)
{
  return(-ESPIPE);
}

void* diva_os_malloc (unsigned long flags, unsigned long size) {
  return (vmalloc (size));
}

void diva_os_free (unsigned long flags, void* ptr) {
  if (ptr) {
    vfree (ptr);
  }
}

static ssize_t
maint_dbg_read(struct file *file, char *buf, size_t count, loff_t * off) {
 return (-ENODEV);
}

static int
maint_dbg_ioctl (struct inode *inode,
             struct file *file,
             uint io_cmd, ulong arg) {
  return (-ENODEV);
}

static unsigned int
maint_dbg_poll(struct file *file, poll_table * wait) {
  return (POLLERR);
}

static int
maint_dbg_open(struct inode *ino, struct file *filep)
{
  filep->private_data = 0;


  return(0);
}

typedef struct _diva_mtpx_um_client {
  _DbgHandle_ hDbg;
} diva_mtpx_um_client_t;

static int
maint_dbg_close(struct inode *ino, struct file *filep)
{

  if (filep->private_data != 0) {
    diva_mtpx_um_client_t* pC = (diva_mtpx_um_client_t*)filep->private_data;

    DbgDeregisterH (&pC->hDbg);
    kfree(filep->private_data);
    filep->private_data = 0;
  }


  return(0);
}

static void dump_dli_data (_DbgHandle_ *hDbg, int dli_name, char* format, ...) {
  va_list ap;

  va_start(ap, format);
  hDbg->dbg_prt (hDbg->id, dli_name, format, ap);
  va_end (ap);
}

static ssize_t
maint_dbg_write(struct file *file, const char *buf, size_t count, loff_t * off)
{
  diva_mtpx_um_client_t* pC = (diva_mtpx_um_client_t*)file->private_data;
  byte cmd;

  if (count < 5)
    return (-EINVAL);

  if (copy_from_user(&cmd, (void *)buf, 1) != 0)
    return(-EFAULT);

  if (pC == 0) {
    if (cmd == DIVA_UM_IDI_TRACE_CMD_REGISTER && count > 8) {
      byte data[128];
      int i;

      if (copy_from_user((void*)&data[0], (void*)(((byte*)buf)+1), MIN(count-1,sizeof(data)-1)))
        return (-EFAULT);

      data[sizeof(data)-1] = 0;
      if ((i = strlen ((char*)&data[4])) == 0 || i >= sizeof(data)-2)
        return (-EINVAL);
      i += 5;

      if ((pC = (diva_mtpx_um_client_t*)kmalloc (sizeof(*pC), GFP_KERNEL)) != 0) {
        memset (pC, 0x00, sizeof(*pC));
        if (DbgRegisterH (&pC->hDbg, &data[4], &data[i], READ_DWORD(data)) == 0) {
          DbgDeregisterH (&pC->hDbg);
          kfree (pC);
          return (-ENOMEM);
        }
        file->private_data = (void*)pC;
        return (count);
      }
      return (-ENOMEM);
    }
    return (-EINVAL);
  }

  if (cmd == DIVA_UM_IDI_TRACE_CMD_WRITE && count > 3) {
    byte data[257+2];
    word dli, length;

    if (copy_from_user((void*)&data[0], (void*)(((byte*)buf)+1), MIN(count-1, sizeof(data)-1)))
      return (-EFAULT);

    dli = READ_WORD(&data[0]);
    switch (dli) {
      case DLI_BLK :
      case DLI_SEND:
      case DLI_RECV:
      case DLI_MXLOG:
      case DLI_LOG :
      case DLI_FTL :
      case DLI_ERR :
      case DLI_TRC :
      case DLI_REG :
      case DLI_MEM :
      case DLI_SPL :
      case DLI_IRP :
      case DLI_TIM :
      case DLI_TAPI:
      case DLI_NDIS:
      case DLI_CONN:
      case DLI_STAT:
      case DLI_PRV0:
      case DLI_PRV1:
      case DLI_PRV2:
      case DLI_PRV3:
        break;

      default:
        return (-EINVAL);
    }

    length = MIN(count - 3, 256);
    data[length+2] = 0;

    dump_dli_data (&pC->hDbg, dli, &data[2], length);

    return (count);
  }

  if (cmd == DIVA_UM_IDI_TRACE_CMD_SET_MASK && count == 5) {
    byte data[4];

    if (copy_from_user((void*)&data[0], (void*)(((byte*)buf)+1), 4))
      return (-EFAULT);

    DbgSetLevelH (&pC->hDbg, READ_DWORD(data));

    return (count);
  }

	return (-EINVAL);
}                                  


static struct file_operations maint_fops =
{
  llseek:         maint_lseek,
  read:           maint_read,
  write:          maint_write,
  poll:           maint_poll,
  ioctl:          maint_ioctl,
  open:           maint_open,
  release:        maint_close,
};

static struct file_operations maint_dbg_fops =
{
  llseek:         maint_lseek,
  read:           maint_dbg_read,
  write:          maint_dbg_write,
  poll:           maint_dbg_poll,
  ioctl:          maint_dbg_ioctl,
  open:           maint_dbg_open,
  release:        maint_dbg_close,
};

static int __init
create_maint_proc(void)
{
  maint_proc_entry = create_proc_entry("maint", S_IFREG | S_IRUGO | S_IWUSR, proc_net_isdn_eicon);
  if (!maint_proc_entry)
    return(0);

  maint_dbg_proc_entry = create_proc_entry("DivasDBG",
                       S_IFREG|S_IRUGO|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH, proc_net_isdn_eicon);

  if (maint_dbg_proc_entry == 0) {
    remove_proc_entry("maint", proc_net_isdn_eicon);
    maint_proc_entry = NULL;
    return (0);
  }

  maint_proc_entry->proc_fops     = &maint_fops;   
  maint_dbg_proc_entry->proc_fops = &maint_dbg_fops;   
  maint_proc_entry->owner     = THIS_MODULE;
  maint_dbg_proc_entry->owner = THIS_MODULE;

  return(1);
}

static void
remove_maint_proc(void)
{
  if (maint_proc_entry) {
    remove_proc_entry("maint", proc_net_isdn_eicon);
    maint_proc_entry = NULL;
  }
  if (maint_dbg_proc_entry != 0) {
    remove_proc_entry("DivasDBG", proc_net_isdn_eicon);
    maint_dbg_proc_entry = NULL;
  }
}

/*
**  DIDD
*/
static void didd_callback (void *context,
                           DESCRIPTOR* adapter,
                           int removal) {
  if (adapter->type == IDI_DADAPTER) {
    printk(KERN_ERR "%s: Change in DAdapter ? Oops ?.\n", DRIVERLNAME);
  } else if (adapter->type == IDI_DIMAINT) {
    if (removal) {
      DbgDeregister();
      memset(&MAdapter, 0, sizeof(MAdapter));
      dprintf = no_printf;
    } else {
      memcpy(&MAdapter, adapter, sizeof(MAdapter));
      dprintf = (DIVA_DI_PRINTF)MAdapter.request;
      DbgRegister("MAINT", DRIVERRELEASE, DBG_DEFAULT);
    }
  } else if ((adapter->type > 0) && (adapter->type < 16)) {
    if (removal) {
      diva_mnt_remove_xdi_adapter(adapter);
    } else {
      diva_mnt_add_xdi_adapter(adapter);
    }
  }
}

static int __init
connect_didd(void)
{
  int x = 0;
  int dadapter = 0;
  IDI_SYNC_REQ req;
  DESCRIPTOR DIDD_Table[MAX_DESCRIPTORS];

  DIVA_DIDD_Read(DIDD_Table, sizeof(DIDD_Table));

  for (x = 0; x < MAX_DESCRIPTORS; x++)
  {
    if (DIDD_Table[x].type == IDI_DADAPTER)
    {  /* DADAPTER found */
      dadapter = 1;
      memcpy(&DAdapter, &DIDD_Table[x], sizeof(DAdapter));
      req.didd_notify.e.Req = 0;
      req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REGISTER_ADAPTER_NOTIFY;
      req.didd_notify.info.callback = didd_callback;
      req.didd_notify.info.context = 0;
      DAdapter.request((ENTITY *)&req);
      if (req.didd_notify.e.Rc != 0xff)
        return(0);
      notify_handle = req.didd_notify.info.handle;
      /* Register MAINT (me) */
      req.didd_add_adapter.e.Req = 0;
      req.didd_add_adapter.e.Rc = IDI_SYNC_REQ_DIDD_ADD_ADAPTER;
      req.didd_add_adapter.info.descriptor = (void *)&MaintDescriptor;
      DAdapter.request((ENTITY *)&req);
      if (req.didd_add_adapter.e.Rc != 0xff)
        return(0);
    } else if ((DIDD_Table[x].type > 0) && (DIDD_Table[x].type < 16)) {
      diva_mnt_add_xdi_adapter(&DIDD_Table[x]);
    }
  }
  return(dadapter);
}

static void __exit
disconnect_didd(void)
{
  IDI_SYNC_REQ req;

  req.didd_notify.e.Req = 0;
  req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REMOVE_ADAPTER_NOTIFY;
  req.didd_notify.info.handle = notify_handle;
  DAdapter.request((ENTITY *)&req);

  req.didd_remove_adapter.e.Req = 0;
  req.didd_remove_adapter.e.Rc = IDI_SYNC_REQ_DIDD_REMOVE_ADAPTER;
  req.didd_remove_adapter.info.p_request = (IDI_CALL)MaintDescriptor.request;
  DAdapter.request((ENTITY *)&req);
}

void diva_maint_wakeup_read (void) {
  wake_up_interruptible (&msgwaitq);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
static void diva_console_write_msg (struct console *con, const char *msg, unsigned int length) {
	if (msg != 0 && *msg != 0 && length > 0) {
		maint_write_system_info (0, msg, (int)length);
	}
}

static struct console diva_maint_console =
	 { flags: CON_ENABLED, write: diva_console_write_msg };

#if FEATURE_NOTIFIER_SUPPORTED
static int diva_notifier_proc (struct notifier_block *self,
                               unsigned long event,
                               void *info);

static struct notifier_block diva_notifier = {
	.notifier_call	= diva_notifier_proc,
	.priority = (INT_MAX-1)
};

static struct notifier_block diva_reboot_notifier = {
	.notifier_call	= diva_notifier_proc,
};

static int diva_notifier_proc (struct notifier_block *self,
                               unsigned long event,
                               void *info) {
	int reboot = (self == &diva_reboot_notifier);

	if (info == 0 || *(char*)info ==0)
		info = (reboot == 0) ? "System message" : "Reboot";

	maint_write_system_info (event, (char*)info, strlen(info));

	if (reboot == 0 ||
			event == SYS_DOWN || event == SYS_RESTART ||
			event == SYS_HALT || event == SYS_POWER_OFF) {
		maint_write_system_info (event, (char*)0, -1);
	}

	return (NOTIFY_DONE);
}
#endif

#endif

static void* __init get_dadapter_request (void) {
	DESCRIPTOR DIDD_Table[MAX_DESCRIPTORS];
	int x = 0;

	DIVA_DIDD_Read(DIDD_Table, sizeof(DIDD_Table));

	for (x = 0; x < MAX_DESCRIPTORS; x++) {
		if (DIDD_Table[x].type == IDI_DADAPTER) {	/* DADAPTER found */
			return (DIDD_Table[x].request);
		}
	}

	return (0);
}

/*
**  Driver Load
*/
static int __init
maint_init(void)
{
  char tmprev[50];
  int ret = 0;
  void* buffer = 0;

  MOD_INC_USE_COUNT;

  do_gettimeofday (&start_time);
  init_waitqueue_head(&msgwaitq);

  printk(KERN_INFO "%s\n", DRIVERNAME);
  printk(KERN_INFO "%s: Rel:%s  Rev:", DRIVERLNAME, DRIVERRELEASE);
  strcpy(tmprev, main_revision);
  printk("%s  Build: %s\n", getrev(tmprev), DIVA_BUILD);
  if (buffer_length < 64) {
    buffer_length = 64;
  }
  if (buffer_length > 7000) {
    buffer_length = 7000;
  }
  buffer_length *= 1024;


  if (diva_dbg_mem) {
    buffer = (void*)diva_dbg_mem;
  } else {
    while ((buffer_length >= (64*1024)) && (!(buffer = kmalloc (buffer_length, GFP_KERNEL)))) {
      buffer_length -= 1024;
    }
  
    if (!buffer) {
      printk(KERN_INFO "%s: Can not alloc trace buffer\n", DRIVERLNAME);
      ret = -EIO;
      goto out;
    }
  }

	if (diva_maint_init_and_connect_to_cfg_lib (buffer,
																							buffer_length,
																							(diva_dbg_mem == 0),
																							get_dadapter_request())) {
    if (!diva_dbg_mem) {
      kfree (buffer);
    }
    ret = -EIO;
    goto out;
  }
  if (!connect_didd()) {
    printk(KERN_ERR "%s: failed to connect to DIDD.\n", DRIVERLNAME);
    diva_maint_finit();
    if (!diva_dbg_mem) {
      kfree (buffer);
    }
    ret = -EIO;
    goto out;
  }
  if (!create_maint_proc()) {
    printk(KERN_ERR "%s: failed to create proc entry.\n", DRIVERLNAME);
    disconnect_didd();
    diva_maint_finit();
    if (!diva_dbg_mem) {
      kfree (buffer);
    }
    ret = -EIO;
    goto out;
  }


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
  if (diva_register_notifier != 0) {
		register_console(&diva_maint_console);
#if FEATURE_NOTIFIER_SUPPORTED
		diva_notifier_registered =
			(notifier_chain_register(&panic_notifier_list, &diva_notifier) == 0);
		diva_reboot_notifier_registered =
			(register_reboot_notifier(&diva_reboot_notifier) == 0);
#endif
  }

	printk(KERN_INFO "%s: trace buffer = %p - %d kBytes, %s%s%s \n",
	       DRIVERLNAME, buffer, (int)(buffer_length / 1024),
	       (diva_dbg_mem == 0) ? "internal" : "external",
         (diva_notifier_registered == 0) ? "" : ", notifier registered",
         (diva_reboot_notifier_registered == 0) ? "" : ", reboot notifier registered");
#else
  printk(KERN_INFO "%s: trace buffer       =%p - %d kBytes, %s \n",
         DRIVERLNAME, buffer, (int)(buffer_length/1024), (diva_dbg_mem == 0) ? "internal" : "external");
#endif


out:
  MOD_DEC_USE_COUNT;
  return(ret);  
}

/*
**  Driver Unload
*/
static void __exit
maint_exit(void)
{
  void* buffer;
  int i = 100;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
	if (diva_register_notifier != 0) {
		unregister_console(&diva_maint_console);
	}

#if FEATURE_NOTIFIER_SUPPORTED
	if (diva_notifier_registered != 0) {
		notifier_chain_unregister(&panic_notifier_list, &diva_notifier);
	}
	if (diva_reboot_notifier_registered != 0) {
		unregister_reboot_notifier(&diva_reboot_notifier);
	}
#endif

#endif

  DbgDeregister();

  while (diva_mnt_shutdown_xdi_adapters () && i--) {
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout(HZ / 10 + 1);
  }

  disconnect_didd();
  remove_maint_proc();

  if ((buffer = diva_maint_finit ())) {
    kfree (buffer);
  }

  printk(KERN_INFO "%s: module unloaded (%d).\n", DRIVERLNAME, i+1);
}

module_init(maint_init);
module_exit(maint_exit);

